/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-iff.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#define AMFF_VERS IFF_ID('V','E','R','S')
#define AMFF_FVER IFF_ID('F','V','E','R')
#define AMFF_HEAD IFF_ID('H','E','A','D')
#define AMFF_BODY IFF_ID('B','O','D','Y')

#define get32(a) be32toh(*(uint32_t*)(a))
#define coord(a) ((double)(a)/16)


typedef struct{
    int32_t x;
    int32_t y;
}point_t;

typedef struct{
    void(*clear_region)(void *userdata, int x1, int y1, int x2, int y2);
    void(*fill_box)(void *userdata, int x1, int y1, int x2, int y2);
    void(*polygon)(void *userdata, int points, const point_t *point);
    void(*curve)(void *userdata, const point_t *point);
    void(*fill_poly)(void *userdata, int points, const point_t *point);
    void(*shape)(void *userdata, int points, const point_t *point);

    void(*setfgpen)(void *userdata, uint32_t r, uint32_t g, uint32_t b);
    void(*setbgpen)(void *userdata, uint32_t r, uint32_t g, uint32_t b);
    void(*setlinewidth)(void *userdata, int width);
}amff_interface_t;


static int
_amff_interpret(const uint32_t *p, size_t len, const amff_interface_t *vt, void *userdata)
{
#define check_args(n) if(args<n) goto error_args
    const uint32_t *end = p+len;
    while(p+2 <= end){
        int opcode;
        int args = p[1];
        if(p+2+args > end)
            return -1;
        opcode = p[0];
        p += 2;
        switch(opcode){
        case 0: /* Fill a rectangle with the BackgroundColor */
            check_args(4);
            vt->clear_region(userdata, p[0], p[1], p[2], p[3]);
            break;
        case 3:
            check_args(4);
            vt->polygon(userdata, 2, (point_t*)p);
            break;
        case 5: /* Write text */
            check_args(3);
            /* text(userdata, p[0], p[1], p[2]); */
            break;
        case 9: /* Draw a filled rectangle */
            check_args(4);
            vt->fill_box(userdata, p[0], p[1], p[2], p[3]);
            break;
        case 11: /* Draw a polygon */
            vt->polygon(userdata, args/2, (point_t*)p);
            break;
        case 14: /* Draw marker */
            check_args(2);
            /* mark(userdata, p[0], p[1]); */
            break;
        case 15: /* Draw beziercurve */
            check_args(8);
            vt->curve(userdata, (point_t*)p);
            break;
        case 16: /* Draw a filled polygon */
            vt->fill_poly(userdata, args/2, (point_t*)p);
            break;
        case 21: /* ??? */
            vt->shape(userdata, args/2, (point_t*)p);
            break;
        case 100: /* Set font with metacoordinates */
            check_args(2);
            /* font(userdata, p[0], p[1]); */
            break;
        case 101: /* Set the foregroundpen */
            check_args(3);
            vt->setfgpen(userdata, p[0], p[1], p[2]);
            break;
        case 102: /* Set the backgroundpen */
            check_args(3);
            vt->setbgpen(userdata, p[0], p[1], p[2]);
        case 103: /* Set drawmode */
            check_args(1);
            /* setdrawmode(userdata, p[0]); */
            break;
        case 104: /* Set linepattern */
            check_args(1);
            /* setlinepat(userdata, p[0]); */
            break;
        case 105: /* Set linewidth */
            check_args(1);
            vt->setlinewidth(userdata, p[0]);
            break;
        case 107: /* Set fillpattern */
            check_args(1);
            /* setfillpat(userdata, p[0]); */
            break;
        case 108: /* Set the size of the marker */
            check_args(1);
            /* setmarksize(userdata, p[0]); */
            break;
        case 109: /* Set type of marker */
            check_args(1);
            /* setmarktype(userdata, p[0]); */
            break;
        default:
            DEBUG_p("unknown opcode: %d (%d args)",opcode,args);
            break;
        }
        p += args;
    }
    return 0;
error_args:
    DEBUG_p("too few arguments to function");
    return -1;
}


static int
_amff_scan_for_size(const uint32_t *p, size_t len, int *width, int *height)
{
    const uint32_t *end = p+len;
    while(p+2 <= end){
        int args = p[1];
        if(p+2+args > end)
            return -1;
        p += 2;
        if(0==p[0] && args>=4){
            *width = p[2];
            *height = p[3];
            return 0;
        }
        p += args;
    }
    return -1;
}

/* --- */

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    uint32_t *data;
    int len;
};


static const char **
_amff_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-amff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_amff_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->scalable_size = 1;
    h->info->threadsafe = 1;

#if 1
    h->info->width = coord(1<<16);
    h->info->height = coord(1<<16);
#endif

    h->data = NULL;
    h->len = 0;

    return h;
}


static void
_amff_free(abydos_plugin_handle_t *h)
{
    if(h->data)
        free(h->data);
    free(h);
}


static int
_handle_amff(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    while(len >= 8){
        size_t chunk_len = get32(data+4);
        if(chunk_len > len-8)
            return -1;
        switch(*(uint32_t*)data){
        case AMFF_VERS:
        case AMFF_FVER:
            break;
        case AMFF_HEAD:
            if(chunk_len < 8)
                return -1;
            h->info->width = get32(data+8);
            h->info->height = get32(data+12);
            DEBUG_p("size: %dx%d", h->info->width, h->info->height);
            break;
        case AMFF_BODY:
            h->len = chunk_len/4;
            h->data = (uint32_t*)malloc(h->len*sizeof(uint32_t));
            DEBUG_p("data size %d (%d words)", (int)chunk_len, h->len);
#if __BYTE_ORDER == __BIG_ENDIAN
            memcpy(h->data, data+8, h->len*sizeof(uint32_t));
#else
            {
                int i;
                for(i=0;i<h->len;++i)
                    h->data[i] = get32(data+8+i*sizeof(uint32_t));
            }
#endif
            {
                int width;
                int height;
                _amff_scan_for_size(h->data, h->len, &width, &height);
                h->info->width = coord(width);
                h->info->height = coord(height);
            }
            break;
        default:
            DEBUG_p("unhandled chunk (%4s)",data);
            break;
        }
        chunk_len += 8+(chunk_len&1);
        data += chunk_len;
        len -= chunk_len;
    }
    return 0;
}


static int
_amff_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    size_t form_len;
    if(len < 12)
        return -1;
    if(*(uint32_t*)data != IFF_FORM || *(uint32_t*)(data+8) != IFF_AMFF)
        return -1;
    form_len = get32(data+4)-4;
    if(form_len > len-12){
        DEBUG_p("truncated data");
        return -1;
    }
    return _handle_amff(h, data+12, form_len);
}


typedef struct{
    abydos_plugin_handle_t *h;
    cairo_t *cr;
    double fg_r;
    double fg_g;
    double fg_b;
    double bg_r;
    double bg_g;
    double bg_b;
}context_t;


static void
_on_clear_region(void *userdata, int x1, int y1, int x2, int y2)
{
    context_t *c = (context_t*)userdata;
    DEBUG_p("%dx%d %dx%d",x1,y1,x2,y2);
#if 0
    if(c->h->info->width <= 0)
        c->h->info->width = x2;
    if(c->h->info->height <= 0)
        c->h->info->height = y2;
#endif
    cairo_set_source_rgb(c->cr, c->bg_r, c->bg_g, c->bg_b);
    cairo_rectangle(c->cr, coord(x1), coord(y1), coord(x2-x1), coord(y2-y1));
    cairo_fill(c->cr);
}


static void
_on_fill_box(void *userdata, int x1, int y1, int x2, int y2)
{
    context_t *c = (context_t*)userdata;
    cairo_set_source_rgb(c->cr, c->fg_r, c->fg_g, c->fg_b);
    cairo_rectangle(c->cr, coord(x1), coord(y1), coord(x2-x1), coord(y2-y1));
    cairo_fill(c->cr);
}


static void
_on_curve(void *userdata, const point_t *p)
{
    context_t *c = (context_t*)userdata;
    cairo_move_to(c->cr, coord(p[0].x), coord(p[0].y));
    cairo_curve_to(c->cr, coord(p[1].x), coord(p[1].y),
                          coord(p[2].x), coord(p[2].y),
                          coord(p[3].x), coord(p[3].y));
    cairo_set_source_rgb(c->cr, c->fg_r, c->fg_g, c->fg_b);
    cairo_stroke(c->cr);
}


static void
_on_polygon(void *userdata, int points, const point_t *p)
{
    context_t *c = (context_t*)userdata;
    int i;
    cairo_move_to(c->cr, coord(p[0].x), coord(p[0].y));
    for(i=1;i<points;++i)
        cairo_line_to(c->cr, coord(p[i].x), coord(p[i].y));
    /* cairo_close_path(c->cr); */
    cairo_set_source_rgb(c->cr, c->fg_r, c->fg_g, c->fg_b);
    cairo_stroke(c->cr);
}


static void
_on_fill_poly(void *userdata, int points, const point_t *p)
{
    context_t *c = (context_t*)userdata;
    int i;
    cairo_move_to(c->cr, coord(p[0].x), coord(p[0].y));
    for(i=1;i<points;++i)
        cairo_line_to(c->cr, coord(p[i].x), coord(p[i].y));
    cairo_set_source_rgb(c->cr, c->fg_r, c->fg_g, c->fg_b);
    cairo_fill(c->cr);
}


static void
_on_shape(void *userdata, int points, const point_t *p)
{
    context_t *c = (context_t*)userdata;
    const point_t *end = p+points;

    cairo_move_to(c->cr, coord(p[0].x), coord(p[0].y));
    ++p;
    for(;;){
        if(p+3 > end)
            break;
        cairo_curve_to(c->cr, coord(p[0].x), coord(p[0].y),
                              coord(p[1].x), coord(p[1].y),
                              coord(p[2].x), coord(p[2].y));
        p += 3;
        if(p>=end)
            break;
        cairo_line_to(c->cr, coord(p[0].x), coord(p[0].y));
        ++p;
    }

    cairo_set_source_rgb(c->cr, c->fg_r, c->fg_g, c->fg_b);
    cairo_fill(c->cr);
}


static void
_on_setfgpen(void *userdata, uint32_t r, uint32_t g, uint32_t b)
{
    context_t *c = (context_t*)userdata;
    c->fg_r = (double)r/0xffffffff;
    c->fg_g = (double)g/0xffffffff;
    c->fg_b = (double)b/0xffffffff;
    DEBUG_p("rgb: %f, %f, %f",c->fg_r,c->fg_g,c->fg_b);
}


static void
_on_setbgpen(void *userdata, uint32_t r, uint32_t g, uint32_t b)
{
    context_t *c = (context_t*)userdata;
    c->bg_r = (double)r/0xffffffff;
    c->bg_g = (double)g/0xffffffff;
    c->bg_b = (double)b/0xffffffff;
    DEBUG_p("rgb: %f, %f, %f",c->bg_r,c->bg_g,c->bg_b);
}


static void
_on_setlinewidth(void* userdata, int width)
{
    context_t *c = (context_t*)userdata;
    DEBUG_p("width: %d",width);
    cairo_set_line_width(c->cr, width);
}


static void
_amff_render(abydos_plugin_handle_t *h,
             cairo_t *cr,
             int page,
             cairo_bool_t *layer_visibility,
             int variant,
             double time,
             int frame)
{
    static const amff_interface_t vt = {
        .clear_region = _on_clear_region,
        .fill_box = _on_fill_box,
        .polygon = _on_polygon,
        .curve = _on_curve,
        .fill_poly = _on_fill_poly,
        .shape = _on_shape,
        .setfgpen = _on_setfgpen,
        .setbgpen = _on_setbgpen,
        .setlinewidth = _on_setlinewidth
    };
    context_t c;
    c.h = h;
    c.cr = cr;
    _amff_interpret(h->data, h->len, &vt, &c);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _amff_get_mime_types,
        .new = _amff_new,
        .free = _amff_free,
        .create_from_data = _amff_create_from_data,
        .render = _amff_render
    };
    return &p;
}
