/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#include <debug.h>

#include "cgmparse.h"
#include "elements.h"


int handle_class1(cgm_context_t *c, int id, cgm_read_t *r)
{
    switch(id){
    case 0x01 : /* Metafile Version (I) */
        DEBUG_p("version: %d", r_get_I(r));
        break;
    case 0x02 : /* Metafile Description (SF) */
        break;
    case 0x03 : /* VDC Type (E) */
        r->vdc_type = r_get_vdc_type(r);
        break;
    case 0x04 : /*Integer Precision*/
        r->int_precision = r_get_int_precision_bytes(r);
        break;
    case 0x05 : /*Real Precision*/
        {
            vdc_type_t type = r_get_vdc_type(r);
            int nI0 = r_get_I(r); /* exponent */
            int nI1 = r_get_I(r); /* mantissa */
            switch( type )
            {
                case 0 :
                    r->real_type = RT_FLOAT;
                    switch ( nI0 )
                    {
                        case 9 :
                            if ( nI1 != 23 )
                                return -1;
                            r->real_size = 4;
                            break;
                        case 12 :
                            if ( nI1 != 52 )
                                return -1;
                            r->real_size = 8;
                            break;
                        default:
                            return -1;
                    }
                    break;
                case 1 :
                    r->real_type = RT_FIXED;
                    if ( nI0 != nI1 )
                        return -1;
                    if ( nI0 == 16 )
                        r->real_size = 4;
                    else if ( nI0 == 32 )
                        r->real_size = 8;
                    else
                        return -1;
                    break;
                default :
                    return -1;
            }
        }
        break;
    case 0x06 : /* Index Precision */
        r->index_precision = r_get_int_precision_bytes(r);
        break;
    case 0x07 : /* Color Precision */
        r->color_precision = r_get_int_precision_bytes(r);
        break;
    case 0x08 : /* Color Index Precision */
        r->color_index_precision = r_get_int_precision_bytes(r);
        break;
    case 0x09 : /* Maximum Colour Index (CI) */
        c->pElement.nColorMaximumIndex = r_get_color_index(r);
        if ( ( c->pElement.nColorMaximumIndex > 256 /*255*/ ) || ( c->pElement.nColorMaximumIndex == 0 ) )
            return -1;
        break;
    case 0x0a : /* Color Value Extent */
        /*
         *  colour value mapping specifier
         *  if the colour model is RGB or CMYK,
         *  minimum colour value (CD)
         *  maximum colour value (CD)
         *  if the colour model is CIELAB, CIELUV, or RGB-related
         *  3 pairs of colour scale and colour offset (6R)
         *  Description:
         */
        break;
    case 0x0b : /* MetaFile Element List */
        break;
    case 0x0c : /* MetaFile Default Replacement */
        DEBUG_p("default replacement");
#if 1
        /* FIXME!! */
#else
        if ( cgm_read_bytes(read) > 1 ){
            std::unique_ptr<uint8_t[]> pBuf(new uint8_t[ cgm_read_bytes(read) ]);
            memcpy( pBuf.get(), read->data, cgm_read_bytes(read) );
            maDefRepList.push_back( std::move(pBuf) );
            maDefRepSizeList.push_back( cgm_read_bytes(read) );
        }
#endif
        break;
    case 0x0d : /* Font List (nSF) */
        while(!cgm_read_eoe(r)){
            int nSize;
            const char *s = r_get_S(r, &nSize);
            if(!s)
                return -1;
            CGMFList_InsertName(&c->pElement.attr.aFontList, s, nSize );
        }
        break;
    case 0x0e : /* Character Set List */
        while(!cgm_read_eoe(r)){
            int nSize;
            const char *s;
            r_get_char_set_type(r);
            s = r_get_S(r, &nSize);
            if(!s)
                return -1;
            CGMFList_InsertCharSet(&c->pElement.attr.aFontList, s, nSize );
        }
        break;
    case 0x0f : /* Character Coding Announcer (E) */
        c->pElement.attr.eCharacterCodingA = r_get_character_coding_a(r);
        break;
    case 0x10 : /* Name Precision */
        break;                   // NS
    case 0x11 : /* Maximum VDC Extent (P, P) */
        break;               // NS
    case 0x12 : /* Segment Priority Extent (I, I) */
        break;          // NS
    case 0x13 : /* Color Model (IX) */
        /*
         *  1: RGB
         *  2: CIELAB
         *  3: CIELUV
         *  4: CMYK
         *  5: RGB-related
         */
        break;                      // NS
    case 0x14 : /* Color Calibration */
        break;                // NS
    case 0x15 : /* Font Properties (n[IX, I, SDR]) */
        break;                  // NS
    case 0x16 : /* Glyph Mapping (IX) */
        break;                    // NS
    case 0x17 : /* Symbol Library List (nSF) */
        break;              // NS
    case 0xfc : /* Inquire Function Support */
        break;
    case 0xfa : /* End Metafile Defaults Replacement */
        break;
    case 0xf8 : /* Set Color Value Desc Extent */
        break;
    default:
        break;
    }
    return 0;
};
