/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#include <debug.h>
#include <loa-buffer.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "chart.h"
#include "cgmparse.h"
#include "elements.h"


static void
_init_path(loa_object_t *shape, loa_buffer_t *vb, loa_path_t *src){
    shape->type = OT_PATH;
    shape->path.point = loa_buffer_get_point(vb, 0);
    shape->path.segment = loa_buffer_get_segment(vb, src->segment);
    shape->path.segment_count = src->segment_count;
}


static int
_calc_center(loa_point_t *center, const loa_point_t *p1, const loa_point_t *p2, const loa_point_t *p3)
{
    double fA = p2->x - p1->x;
    double fB = p2->y - p1->y;
    double fC = p3->x - p1->x;
    double fD = p3->y - p1->y;

    double fE = fA * ( p1->x + p2->x ) + fB * ( p1->y + p2->y );
    double fF = fC * ( p1->x + p3->x ) + fD * ( p1->y + p3->y );

    double fG = 2.0 * ( fA * ( p3->y - p2->y ) - fB * ( p3->x - p2->x ) );

    if(0 == fG)
        return 0;

    center->x = ( fD * fE - fB * fF ) / fG;
    center->y = ( fA * fF - fC * fE ) / fG;
    return 1;
}


static double
_calc_orientation(const loa_point_t *rCenter, const loa_point_t *rPoint )
{
    return atan2((rPoint->y - rCenter->y), (rPoint->x - rCenter->x));
}


static double
_calc_radius(const loa_point_t *center, const loa_point_t *p)
{
    return sqrt( pow( ( p->x - center->x ), 2 ) + pow( ( p->y - center->y ), 2 ) ) ;
}


static void
_calc_ellipse_matrix(loa_matrix_t *m, const arc_t *arc)
{
    loa_matrix_set_identity(m);
    loa_matrix_scale(m, arc->size.x, arc->size.y);
    loa_matrix_rotate(m, arc->orientation);
}


#if 0
static void
_switch_angles(arc_t *arc)
{
    double nTemp = arc->startangle;
    arc->startangle = arc->endangle;
    arc->endangle = nTemp;
}
#endif


static double
_get_angle(cgm_read_t *r, const loa_matrix_t *m)
{
    loa_point_t p;
    r_get_P(r, &p);
    if(m)
        loa_matrix_apply_point(m, &p);
    return atan2(p.y, p.x);
}


static void
_get_ellipse(cgm_read_t *r, arc_t *arc)
{
    loa_point_t aPoint1;
    loa_point_t aPoint2;
    double fRot1;
    double fRot2;
    r_get_P(r, &arc->pos);
    r_get_P(r, &aPoint1);
    r_get_P(r, &aPoint2);
    fRot1 = _calc_orientation(&arc->pos, &aPoint1);
    fRot2 = _calc_orientation(&arc->pos, &aPoint2);
    arc->orientation = fRot1;
    arc->size.x = _calc_radius(&arc->pos, &aPoint1);
    arc->size.y = _calc_radius(&arc->pos, &aPoint2);

    if ( fRot1 > fRot2 ){
        arc->negative = ( fRot1 - fRot2 ) < M_PI;
    }else{
        arc->negative = ( fRot2 - fRot1 ) > M_PI;
    }
}


int handle_class4(cgm_context_t *c, int id, cgm_read_t *r)
{
    loa_object_t shape;
    if ( c->mbFirstOutPut )
        c->vt->FirstOutPut(c->userdata, c->width, c->height, c->mnVDCXadd, c->mnVDCYadd);

#if 0
    if ( mpBitmapInUse && ( mnElementID != 9 ) )    // process existed graphic
    {                                               // because there are now no pending bitmap actions
        //CGMBitmapDescriptor* pBmpDesc = mpBitmapInUse->GetBitmap();
        // do anything with the bitmap
        /* FIXME!! */
        //vt->DrawBitmap(userdata, pBmpDesc );
        mpBitmapInUse.reset();
    }
#endif

    if( 1 /* ( mpChart == nullptr ) || mpChart->IsAnnotation() */ ){
    }else{
        return 0;
    }

    switch(id){
    case 0x01 : /* PolyLine (nP)*/
        {
            loa_path_t path;
            //path_t path2;
            path.segment = loa_buffer_segment_count(&c->vb);
            loa_point_t *p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
            r_get_P(r, p);
            while ( !cgm_read_eoe(r) ){
                p = loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                r_get_P(r, p);
            }
            path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
            _init_path(&shape, &c->vb, &path);
            if(c->mbFigure)
                c->vt->RegPolyLine(c->userdata, &shape.path, false);
            else
                c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x02 : /*Disjoint PolyLine (nP) */
        {
#if 1
            loa_path_t path;
            path.segment = loa_buffer_segment_count(&c->vb);
            while ( !cgm_read_eoe(r) ){
                loa_point_t *p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                r_get_P(r, p);
                p = loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                r_get_P(r, p);
            }
            path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
            _init_path(&shape, &c->vb, &path);
            if(c->mbFigure){
                c->vt->RegPolyLine(c->userdata, &shape.path, false );
            }else{
                c->vt->BeginGroup(c->userdata);
                c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
                c->vt->EndGroup(c->userdata);
            }
#else
            uint16_t nPoints = sal::static_int_cast< uint16_t >(
                mnElementSize / ImplGetPointSize());
            if ( ! ( nPoints & 1 ) )
            {
                nPoints >>= 1;
                //point_t  aFloatPoint;
                if ( mbFigure )
                {
                    loa_path_t path;
                    path.segment = loa_buffer_segment_count(&c->vb);
                    loa_point_t *p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                    r_get_point(r, p[0]);
                    for ( int i = 1; i < nPoints; i++){
                        p = loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                        r_get_point(r, p[0]);
                    }
                    path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
                    vt->RegPolyLine(userdata, &path, false );
                }
                else
                {
                    vt->BeginGroup(userdata);
                    loa_path_t path;
                    path.segment = loa_buffer_segment_count(&c->vb);
                    loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                    loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                    point_t *p = loa_buffer_get_point(&c->vb, loa_buffer_get_segment(&c->vb, path.segment)->point);
                    for ( uint16_t i = 0; i < nPoints; i++ )
                    {
                        r_get_point(r, p[0]);
                        r_get_point(r, p[1]);
                        vt->stroke_path(userdata, &c->pElement.aLineBundle, &path );
                    }
                    vt->EndGroup(userdata);
                }
            }
#endif
        }
        break;
    case 0x03 : /* PolyMarker (nP) */
        break;
    case 0x04 : /* Text (P, E, S) */
        {
            rect_t rect;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            rect.size.x = rect.size.y = 0;
            r_get_P(r, &rect.pos);
            FinalFlag f = r_get_final_flag(r);
            int nSize;
            const char *s = r_get_S(r, &nSize);
            char *str = strndup(s, nSize);
            c->vt->DrawText(c->userdata, &rect, &c->pElement.attr, str, f);
            free(str);
        }
        break;
    case 0x05 : /* Restricted Text (2VDC, P, E, S) */
        {
            rect_t rect;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            r_get_P(r, &rect.size);
            r_get_P(r, &rect.pos);
            FinalFlag f = r_get_final_flag(r);
            int nSize;
            const char *s = r_get_S(r, &nSize);
            char *str = strndup(s, nSize);
            c->vt->DrawText(c->userdata, &rect, &c->pElement.attr, str, f);
            free(str);
        }
        break;
    case 0x06 : /* Append Text (E, S) */
        {
            FinalFlag f = r_get_final_flag(r); // nType
            int nSize;
            const char *s = r_get_S(r, &nSize);
            char *str = strndup(s, nSize);
            c->vt->AppendText(c->userdata, &c->pElement.attr, str, f);
            free(str);
        }
        break;
    case 0x07 : /* Polygon (nP) */
        {
            loa_path_t path;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            path.segment = loa_buffer_segment_count(&c->vb);
            loa_point_t *p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
            r_get_P(r, p);
            while ( !cgm_read_eoe(r) ){
                p = loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                r_get_P(r, p);
            }
            loa_buffer_push_segment(&c->vb, &path, LOA_CLOSE);
            path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
            _init_path(&shape, &c->vb, &path);
            c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x08 : /* Polygon Set (n[P, E]) */
        {
            loa_path_t path;
            edge_flag_t nEdgeFlag = EF_INVISIBLE;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            loa_buffer_clear(&c->vb);
            path.segment = loa_buffer_segment_count(&c->vb);
            //int first_point = loa_buffer_point_count(&c->vb);

            while ( !cgm_read_eoe(r) )
            {
                loa_point_t *p;
                if(nEdgeFlag & EF_VISIBLE_BIT){
                    p = loa_buffer_push_segment(&c->vb, &path, LOA_LINE);
                }else{
                    p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                }
                r_get_P(r, p);
                nEdgeFlag = r_get_edge_flag(r);
                if ( nEdgeFlag & EF_CLOSE_BIT )
                {
                    loa_buffer_push_segment(&c->vb, &path, LOA_CLOSE);
                    nEdgeFlag = EF_INVISIBLE;
                }
            }
            //loa_buffer_push_segment(&c->vb, &path, LOA_CLOSE);
            path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
            _init_path(&shape, &c->vb, &path);
            c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x09 : /* Cell Array (3P, 2I, nCO) */
        {
            loa_point_t p[3];
            int width;
            int height;
            int old_prec;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            r_get_P(r, &p[0]);
            r_get_P(r, &p[1]);
            r_get_P(r, &p[2]);
            width = r_get_I(r);
            height = r_get_I(r);
            old_prec = r->int_precision;
            r->int_precision = r_get_int_precision_bytes(r);

#if 1
            nile_t ni;
            nile_init(&ni);
            nile_ensure_indexed(&ni, width, height);
            int stride = nile_get_stride(&ni);
            DEBUG_p("bitmap size: %dx%d", width, height);
            for(int y=0;y<height;++y){
                int x;
                for(x=0;x<width;++x){
                    ni.pixels[x + y*stride] = r_get_color_index(r);
                }
            }
            ni.palette = (uint8_t*)c->pElement.aColorTable;
            ni.colors = c->pElement.nColorMaximumIndex;
            ni.bpc = 4;
#if __BYTE_ORDER == __BIG_ENDIAN
            ni.format = NILE_FORMAT_x8r8g8b8
#else
            ni.format = NILE_FORMAT_b8g8r8;
#endif
            c->vt->DrawBitmap(c->userdata, p, &ni);
            nile_done(&ni);
#else
            if ( mpBitmapInUse )
            {
                std::unique_ptr<CGMBitmap> xBmpDesc(mpBitmapInUse->GetNext());
                if (xBmpDesc) // we possibly get a bitmap back which does not fit to
                {               // to the previous -> we need to delete this one too
                    vt->DrawBitmap(userdata, xBmpDesc->GetBitmap());
                }
            }
            else
            {
                mpBitmapInUse.reset( new CGMBitmap( *this ) );
            }
#endif
            r->int_precision = old_prec;
        }
        break;
    case 0x0a : /* Generalized Drawing Primitive (I, nP, D) */
        //r_get_int(r);  //-Wall is this needed
        //r_get_uint(r); //-Wall is this needed
        break;
    case 0x0b : /* Rectangle (2P) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);

        shape.type = OT_RECT;
        r_get_rect(r, &shape.rect);
        c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        break;
    case 0x0c : /* Circle (P, VDC) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);

        shape.type = OT_ARC;
        r_get_P(r, &shape.arc.pos);
        shape.arc.size.x = shape.arc.size.y = r_get_VDC(r); // r_get_xpos(r);
        shape.arc.startangle = 0;
        shape.arc.endangle = 2*M_PI;
        shape.arc.close_type = CT_CHORD;
        shape.arc.negative = 0;
        c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        break;
    case 0x0d : /* Circular Arc 3 Point (3P, E) */
        {
            loa_point_t aStartingPoint;
            loa_point_t aIntermediatePoint;
            loa_point_t aEndingPoint;

            shape.type = OT_ARC;
            r_get_P(r, &aStartingPoint);
            //vt->DrawEllipse(userdata, aStartingPoint, {25,25}, 0);
            r_get_P(r, &aIntermediatePoint);
            //vt->DrawEllipse(userdata, aIntermediatePoint, {50,50}, 0);
            r_get_P(r, &aEndingPoint);
            //vt->DrawEllipse(userdata, aEndingPoint, {25,25}, 0);

            if(_calc_center(&shape.arc.pos, &aStartingPoint, &aIntermediatePoint, &aEndingPoint)){
                double fInterAngle;
                shape.arc.startangle = _calc_orientation(&shape.arc.pos, &aStartingPoint );
                fInterAngle = _calc_orientation(&shape.arc.pos, &aIntermediatePoint );
                shape.arc.endangle = _calc_orientation(&shape.arc.pos, &aEndingPoint );
                shape.arc.negative = (shape.arc.startangle > fInterAngle) ^
                                     (shape.arc.startangle > shape.arc.endangle) ^
                                     (fInterAngle > shape.arc.endangle);

                shape.arc.size.x = shape.arc.size.y = _calc_radius(&shape.arc.pos, &aStartingPoint);
                shape.arc.orientation = 0;

                if(c->mbFigure){
#if 1
                    /* FIXME!! */
#else
                    tools::Rectangle aBoundingBox(aCenterPoint.x - fRadius, aCenterPoint.y - fRadius);
                    aBoundingBox.SaturatingSetSize(Size(2 * fRadius, 2 * fRadius));
                    tools::Polygon aPolygon( aBoundingBox, Point( static_cast<long>(aStartingPoint.x), static_cast<long>(aStartingPoint.y) ) ,Point( static_cast<long>(aEndingPoint.x), static_cast<long>(aEndingPoint.y) ), PolyStyle::Arc );
                    if ( nSwitch )
                        vt->RegPolyLine(userdata, aPolygon, true );
                    else
                        vt->RegPolyLine(userdata, aPolygon, false );
#endif
                }
                else
                {
                    c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
                }
            }
        }
        break;
    case 0x0e : /* Circular Arc 3 Point Close (3P, E) */
        {
            loa_point_t aStartingPoint;
            loa_point_t aIntermediatePoint;
            loa_point_t aEndingPoint;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            shape.type = OT_ARC;
            r_get_P(r, &aStartingPoint);
            //vt->DrawEllipse(userdata, aStartingPoint, {25,25}, 0);
            r_get_P(r, &aIntermediatePoint);
            //vt->DrawEllipse(userdata, aIntermediatePoint, {50,50}, 0);
            r_get_P(r, &aEndingPoint);
            //vt->DrawEllipse(userdata, aEndingPoint, {25,25}, 0);

            if(_calc_center(&shape.arc.pos, &aStartingPoint, &aIntermediatePoint, &aEndingPoint)){
                double fInterAngle;
                shape.arc.startangle = _calc_orientation(&shape.arc.pos, &aStartingPoint );
                fInterAngle = _calc_orientation(&shape.arc.pos, &aIntermediatePoint );
                shape.arc.endangle = _calc_orientation(&shape.arc.pos, &aEndingPoint );
                shape.arc.negative = (shape.arc.startangle > fInterAngle) ^
                                     (shape.arc.startangle > shape.arc.endangle) ^
                                     (fInterAngle > shape.arc.endangle);
                shape.arc.size.x = shape.arc.size.y = _calc_radius(&shape.arc.pos, &aStartingPoint);
                shape.arc.orientation = 0;
                shape.arc.close_type = r_get_close_type(r);
                c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
            }
        }
        break;
    case 0x0f : /* Circular Arc Centre (P, 4VDC, VDC) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);

        shape.type = OT_ARC;
        r_get_P(r, &shape.arc.pos);
        shape.arc.startangle = _get_angle(r, NULL);
        shape.arc.endangle = _get_angle(r, NULL);
        shape.arc.size.x = shape.arc.size.y = r_get_VDC(r);

        if(c->mbFigure){
#if 0
            /* FIXME!! */
            tools::Rectangle aBoundingBox(aCenter.x - aRadius.x, aCenter.y - aRadius.x);
            aBoundingBox.SaturatingSetSize(Size(2 * aRadius.x, 2 * aRadius.x));
            tools::Polygon aPolygon( aBoundingBox,
                Point( static_cast<long>(vector[ 0 ]), static_cast<long>(vector[ 1 ]) ),
                Point( static_cast<long>(vector[ 2 ]), static_cast<long>(vector[ 3 ]) ), PolyStyle::Arc );
            //vt->RegPolyLine(userdata, aPolygon, false );
#endif
        }else{
            c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x10 : /* Circular Arc Centre Close (P, 4VDC, VDC, E) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);

        shape.type = OT_ARC;
        r_get_P(r, &shape.arc.pos);
        shape.arc.startangle = _get_angle(r, NULL);
        shape.arc.endangle = _get_angle(r, NULL);
        shape.arc.size.x = shape.arc.size.y = r_get_VDC(r);
        shape.arc.orientation = 0;
        shape.arc.close_type = r_get_close_type(r);
        c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        break;
    case 0x11 : /* Ellipse (P, P, P) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);

        shape.type = OT_ARC;
        _get_ellipse(r, &shape.arc) ;
        shape.arc.startangle = 0;
        shape.arc.endangle = 2*M_PI;
        shape.arc.close_type = CT_CHORD;
        shape.arc.negative = 0;
        c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        break;
    case 0x12 : /* Elliptical Arc (P, P, P, 4VDC) */
        {
            loa_matrix_t m;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            shape.type = OT_ARC;
            _get_ellipse(r, &shape.arc);
            //arc.orientation = 0;
            _calc_ellipse_matrix(&m, &shape.arc);
            shape.arc.startangle = _get_angle(r, &m);
            shape.arc.endangle = _get_angle(r, &m);
            c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x13 : /* Elliptical Arc Close (P, P, P, 4VDC, E) */
        {
            loa_matrix_t m;

            if(c->mbFigure)
                c->vt->CloseRegion(c->userdata);

            shape.type = OT_ARC;
            _get_ellipse(r, &shape.arc);
            _calc_ellipse_matrix(&m, &shape.arc);
            shape.arc.startangle = _get_angle(r, &m);
            shape.arc.endangle = _get_angle(r, &m);
            shape.arc.close_type = r_get_close_type(r);
            c->vt->fill_shape(c->userdata, &c->pElement.attr, &shape);
        }
        break;
    case 0x14 : /* Circular Arc Centre Reversed (P, 4VDC, VDC) */
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x15 : /* Connection Edge () */                      // NS
//      if(c->mbFigure)
//          mpOutAct->CloseRegion();
        break;
    case 0x16 : /* Hyperbolic Arc (P, P, P, 2VDC, 2VDC) */                      // NS
        DEBUG_p("hyperbolic arc not supported");
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x17 : /* Parabolic Arc (P, P, P) */                      // NS
        DEBUG_p("parabolic arc not supported");
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x18 : /* Non Uniform B-Spline (I, I, nP, (k+n)R, R, R) */              // NS
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x19 : /* Non Uniform Rational B-Spline (I, I, nP (k+n)R, R, R, nR) */      // NS
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x1a : /* Polybezier (IX, [4nP | (3n+1)P]) */
        {
            uint32_t nOrder = r_get_I(r);
#if 1
            loa_path_t path;
            //path_t path2;
            path.segment = loa_buffer_segment_count(&c->vb);
            loa_point_t *p;
            if ( nOrder & 4 )
            {
                p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                r_get_P(r, p);
                while ( !cgm_read_eoe(r) )
                {
                    p = loa_buffer_push_segment(&c->vb, &path, LOA_CURVE);
                    r_get_P(r, &p[0]);
                    r_get_P(r, &p[1]);
                    r_get_P(r, &p[2]);
                }
            }
            else
            {
                while ( !cgm_read_eoe(r) )
                {
                    p = loa_buffer_push_segment(&c->vb, &path, LOA_GOTO);
                    r_get_P(r, p);
                    p = loa_buffer_push_segment(&c->vb, &path, LOA_CURVE);
                    r_get_P(r, &p[0]);
                    r_get_P(r, &p[1]);
                    r_get_P(r, &p[2]);
                }
            }
            loa_buffer_push_segment(&c->vb, &path, LOA_CLOSE);
            path.segment_count = loa_buffer_segment_count(&c->vb) - path.segment;
            _init_path(&shape, &c->vb, &path);
            if(c->mbFigure)
                c->vt->RegPolyLine(c->userdata, &shape.path, false );
            else
                c->vt->stroke_shape(c->userdata, &c->pElement.attr, &shape);
#endif
        }
        break;
    case 0x1b : /*Polysymbol (IX, nP) */                          // NS
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x1c : /*Bitonal Tile (IX, I, CO, CO, SDR, BS) */                      // NS
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        /*
         *  0: null background
         *  1: null foreground
         *  2: T6
         *  3: T4 1-dimensional
         *  4: T4 2-dimensional
         *  5: bitmap (uncompressed)
         */
        break;
    case 0x1d : /*Tile (IX, I, ?, SDR, BS) */                              // NS
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0x1e : /*Insert Object*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xff : /*Polybezier*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xfe : /*Sharp Polybezier*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xfd : /*Polyspline*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xfc : /*Rounded Rectangle*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xfb : /*Begin Cell Array*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xfa : /*End Cell Array*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xf9 : /*Insert File*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xf8 : /*Block Text*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xf7 : /*Variable Width Polyline*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xf6 : /*Elliptical Arc 3 Point*/
        if(c->mbFigure)
            c->vt->CloseRegion(c->userdata);
        break;
    case 0xf1 : /*Hyperlink Definition  */
        break;
    default:
        break;
    }
    return 0;
};
