/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#include <debug.h>

#include "cgmparse.h"
#include "elements.h"


int handle_class5(cgm_context_t *c, int id, cgm_read_t *r)
{
    switch(id){
    case 0x01 : /* Line Bundle Index (IX) */
        {
            int idx = r_get_IX(r);
            LineBundle *bn = (LineBundle*)cgm_get_bundle(&c->pElement.attr.line_list, idx);
            if(!bn){
                bn = (LineBundle*)loa_array_append(&c->pElement.attr.line_list, 1);
                bn->base.idx = idx;
                bn->base.color = select_line_bundle(&c->pElement.attr, ASF_LINECOLOR)->base.color;
                bn->type = a_get_line_type(&c->pElement.attr);
                bn->width = a_get_line_width(&c->pElement.attr);
            }
            c->pElement.attr.line_bundled = loa_array_index(&c->pElement.attr.line_list, bn);
        }
        break;
    case 0x02 : /* Line Type (IX) */
        select_line_bundle(&c->pElement.attr, ASF_LINETYPE)->type = (line_type_t)r_get_IX(r);
        break;
    case 0x03 : /* Line Width (SS) */
        select_line_bundle(&c->pElement.attr, ASF_LINEWIDTH)->width = r_get_V(r, c->pElement.attr.eLineWidthSpecMode);
        break;
    case 0x04 : /* Line Color (CO) */
        select_line_bundle(&c->pElement.attr, ASF_LINECOLOR)->base.color = r_get_color(r, &c->pElement, 0);
        break;
    case 0x05 : /* Marker Bundle Index (IX) */
        {
            int idx = r_get_IX(r);
            MarkerBundle *bn = (MarkerBundle*)cgm_get_bundle(&c->pElement.attr.marker_list, idx);
            if(!bn){
                bn = (MarkerBundle*)loa_array_append(&c->pElement.attr.marker_list, 1);
                bn->base.idx = idx;
                bn->base.color = select_marker_bundle(&c->pElement.attr, ASF_MARKERCOLOR)->base.color;
                bn->eMarkerType = select_marker_bundle(&c->pElement.attr, ASF_MARKERTYPE)->eMarkerType;
                bn->nMarkerSize = select_marker_bundle(&c->pElement.attr, ASF_MARKERSIZE)->nMarkerSize;
            }
            c->pElement.attr.marker_bundled = loa_array_index(&c->pElement.attr.marker_list, bn);
        }
        break;
    case 0x06 : /* Marker Type (IX) */
        select_marker_bundle(&c->pElement.attr, ASF_MARKERTYPE)->eMarkerType = r_get_marker_type(r);
        break;
    case 0x07 : /* Marker Size (SS) */
        select_marker_bundle(&c->pElement.attr, ASF_MARKERSIZE)->nMarkerSize = r_get_V(r, c->pElement.attr.eMarkerSizeSpecMode);
        break;
    case 0x08 : /* Marker Color (CO) */
        select_marker_bundle(&c->pElement.attr, ASF_MARKERCOLOR)->base.color = r_get_color(r, &c->pElement, 0);
        break;
    case 0x09 : /* Text Bundle Index (IX) */
        {
            int idx = r_get_IX(r);
            TextBundle *bn = (TextBundle*)cgm_get_bundle(&c->pElement.attr.text_list, idx);
            if(!bn){
                bn = (TextBundle*)loa_array_append(&c->pElement.attr.text_list, 1);
                bn->base.idx = idx;
                bn->base.color = a_get_text_color(&c->pElement.attr);
                bn->nTextFontIndex = select_text_bundle(&c->pElement.attr, ASF_TEXTFONTINDEX)->nTextFontIndex;
                bn->eTextPrecision = select_text_bundle(&c->pElement.attr, ASF_TEXTPRECISION)->eTextPrecision;
                bn->nCharacterExpansion = select_text_bundle(&c->pElement.attr, ASF_CHARACTEREXPANSION)->nCharacterExpansion;
                bn->nCharacterSpacing = select_text_bundle(&c->pElement.attr, ASF_CHARACTERSPACING)->nCharacterSpacing;
            }
            c->pElement.attr.text_bundled = loa_array_index(&c->pElement.attr.text_list, bn);
        }
        break;
    case 0x0a : /* Text Font Index (IX) */
        select_text_bundle(&c->pElement.attr, ASF_TEXTFONTINDEX)->nTextFontIndex = r_get_IX(r);
        break;
    case 0x0b : /* Text Precision (E) */
        select_text_bundle(&c->pElement.attr, ASF_TEXTPRECISION)->eTextPrecision = r_get_text_precision(r);
        break;
    case 0x0c : /* Character Expansion Factor (R) */
        select_text_bundle(&c->pElement.attr, ASF_CHARACTEREXPANSION)->nCharacterExpansion = r_get_R(r);
        break;
    case 0x0d : /* Character Spacing (R) */
        select_text_bundle(&c->pElement.attr, ASF_CHARACTERSPACING)->nCharacterSpacing = r_get_R(r);
        break;
    case 0x0e : /* Text Color (CO) */
        select_text_bundle(&c->pElement.attr, ASF_TEXTCOLOR)->base.color = r_get_color(r, &c->pElement, 0);
        break;
    case 0x0f : /* Character Height (VDC) */
#if 1
        c->pElement.attr.nCharacterHeight = r_get_VDC(r);
#else
        c->pElement.nCharacterHeight = r_get_vdc(r);
        cgm_map_double(c, r, &c->pElement.nCharacterHeight );
        //c->pElement.nCharacterHeight /= 18.0;
#endif
        break;
    case 0x10 : /* Character Orientation (VDC, VDC, VDC, VDC) */
        {
            int i;
            for(i=0;i<4;++i)
                c->pElement.attr.nCharacterOrientation[i] = r_get_VDC(r);
        }
        break;
    case 0x11 : /* Text Path (E) */
        c->pElement.attr.eTextPath = r_get_text_path(r);
        break;
    case 0x12 : /* Text Alignment (E, E, R, R) */
        c->pElement.attr.eTextAlignmentH = r_get_text_alignment_h(r);
        c->pElement.attr.eTextAlignmentV = r_get_text_alignment_v(r);
        c->pElement.attr.nTextAlignmentHCont = r_get_R(r);
        c->pElement.attr.nTextAlignmentVCont = r_get_R(r);
        break;
    case 0x13 : /* Character Set Index (IX) */
        c->pElement.attr.nCharacterSetIndex = r_get_IX(r);
        break;
    case 0x14 : /* Alternate Character Set Index (IX) */
        c->pElement.attr.nAlternateCharacterSetIndex = r_get_IX(r);
        break;
    case 0x15 : /* Fill Bundle Index (IX) */
        {
            int idx = r_get_IX(r);
            FillBundle *bn = (FillBundle*)cgm_get_bundle(&c->pElement.attr.fill_list, idx);
            if(!bn){
                bn = (FillBundle*)loa_array_append(&c->pElement.attr.fill_list, 1);
                bn->base.idx = idx;
                bn->base.color = a_get_fill_color(&c->pElement.attr);
                bn->eFillInteriorStyle = select_fill_bundle(&c->pElement.attr, ASF_FILLINTERIORSTYLE)->eFillInteriorStyle;
                bn->nFillPatternIndex = select_fill_bundle(&c->pElement.attr, ASF_PATTERNINDEX)->nFillPatternIndex;
                bn->nFillHatchIndex = select_fill_bundle(&c->pElement.attr, ASF_HATCHINDEX)->nFillHatchIndex;
            }
            c->pElement.attr.fill_bundled = loa_array_index(&c->pElement.attr.fill_list, bn);
        }
        break;
    case 0x16 : /* Fill Interior Style (E) */
        select_fill_bundle(&c->pElement.attr, ASF_FILLINTERIORSTYLE)->eFillInteriorStyle = r_get_interior_style(r);
        break;
    case 0x17 : /* Fill Color (CO) */
        select_fill_bundle(&c->pElement.attr, ASF_FILLCOLOR)->base.color = r_get_color(r, &c->pElement, 0);
        break;
    case 0x18 : /* Fill Hatch Index (IX) */
        select_fill_bundle(&c->pElement.attr, ASF_HATCHINDEX)->nFillHatchIndex = r_get_IX(r);
        break;
    case 0x19 : /* Fill Pattern Index (IX) */
        select_fill_bundle(&c->pElement.attr, ASF_PATTERNINDEX)->nFillPatternIndex = r_get_IX(r);
        break;
    case 0x1a : /* Edge Bundle Index (IX) */
        {
            int idx = r_get_IX(r);
            LineBundle *bn = (LineBundle*)cgm_get_bundle(&c->pElement.attr.edge_list, idx);
            if(!bn){
                bn = loa_array_append(&c->pElement.attr.edge_list, 1);
                bn->base.idx = idx;
                bn->base.color = a_get_edge_color(&c->pElement.attr);
                bn->type = a_get_edge_type(&c->pElement.attr);
                bn->width = a_get_edge_width(&c->pElement.attr);
            }
            c->pElement.attr.edge_bundled = loa_array_index(&c->pElement.attr.edge_list, bn);
        }
        break;
    case 0x1b : /* Edge Type (IX) */
        select_edge_bundle(&c->pElement.attr, ASF_EDGETYPE)->type = (line_type_t)r_get_IX(r);
        break;
    case 0x1c : /* Edge Width (SS) */
        select_edge_bundle(&c->pElement.attr, ASF_EDGEWIDTH)->width = r_get_V(r, c->pElement.attr.eEdgeWidthSpecMode);
        break;
    case 0x1d : /* Edge Color (CO) */
        select_edge_bundle(&c->pElement.attr, ASF_EDGECOLOR)->base.color = r_get_color(r, &c->pElement, 0);
        break;
    case 0x1e : /* Edge Visibility (E) */
        c->pElement.attr.eEdgeVisibility = r_get_edge_visibility(r);
        break;
    case 0x1f : /* Fill Reference Point (P) */
        r_get_P(r, &c->pElement.attr.aFillRefPoint);
        break;
    case 0x20 : /* Pattern Table (IX, 2I, ?, (nx*ny)CO) */
        break;
    case 0x21 : /* Pattern Size (4SS) */
        break;
    case 0x22 : /* Color Table (CI, nCD) */
        {
            int nColorStartIndex = r_get_color_index(r);
            int nIndex;
            for ( nIndex = nColorStartIndex; nIndex < 256; nIndex++ ){
                if ( cgm_read_eoe(r) )
                    break;
                if ( nIndex > c->pElement.nLatestColorMaximumIndex )
                    c->pElement.nLatestColorMaximumIndex = nIndex;
                c->pElement.aLatestColorTable[ nIndex ] = r_get_color(r, &c->pElement, 1);
                if ( !c->pElement.aColorTableEntryIs[ nIndex ] )
                {
                    c->pElement.aColorTableEntryIs[ nIndex ] = 1;
                    c->pElement.aColorTable[ nIndex ] = c->pElement.aLatestColorTable[ nIndex ];
                }
            }
            c->pElement.nColorMaximumIndex = c->pElement.nLatestColorMaximumIndex;
        }
        break;
    case 0x23 : /*Aspect Source Flags (n[E, E]) */
        while(!cgm_read_eoe(r)){
            int mask = r_get_asf_mask(r);
            switch(r_get_asf_value(r)){
            case ASF_INDIVIDUAL:
                c->pElement.attr.nAspectSourceFlags &= ~mask;
                break;
            case ASF_BUNDLED:
                c->pElement.attr.nAspectSourceFlags |= mask;
                break;
            default:
                return -1;
            }
        }
        break;
    case 0x24 : /* Pick Identifier (N) */
        break;
    case 0x25 : /* Line Cap (IX, IX) */
        c->pElement.attr.eLineCapType = (LineCapType)r_get_IX(r);
        /* FIXME!! read dash cap indicator */
        break;
    case 0x26 : /* Line Join (IX) */
        c->pElement.attr.eLineJoinType = (LineJoinType)r_get_IX(r);
        break;
    case 0x27 : /* Line Type Continuation (IX) */
        break;           // NS
    case 0x28 : /* Line Type Initial Offset (R) */
        break;         // NS
    case 0x29 : /* Text Score Type (n[IX, E]) */
        break;
    case 0x2a : /* Restricted Text Type (IX) */
        break;
    case 0x2b : /* Interpolated interior (IX, 2nSS, I, mR, [(m+1)CO | 3CO]) */
        break;
    case 0x2c : /* Edge Cap (IX, IX) */
        break;                         // NS
    case 0x2d : /* Edge Join (IX) */
        break;
    case 0x2e : /* Edge Type Continuation (IX) */
        break;           // NS
    case 0x2f : /* Edge Type Initial Offset (R) */
        break;         // NS
    case 0x30 : /* Symbol Library Index (IX) */
        break;             // NS
    case 0x31 : /* Symbol Color (CO) */
        break;                     // NS
    case 0x32 : /* Symbol Size (E, VDC, VDC) */
        break;                      // NS
    case 0x33 : /* Symbol Orientation (VDC, VDC, VDC, VDC) */
        break;               // NS
    case 0x50 : /* Block Text Region Margins */
        break;
    case 0x51 : /* Block Text Region Expansion */
        break;
    case 0x52 : /* Block Text Region Anchor */
        break;
    case 0x53 : /* Block Text Paragraph Horizontal Alignment */
        break;
    case 0x54 : /* Block Text Paragraph Vertical Alignment */
        break;
    case 0x55 : /* Block Text Line Flow */
        break;
    case 0x60 : /* Block Text Paragraph Spacing */
        break;
    case 0x61 : /* Block Text Paragraph Indent */
        break;
    case 0x62 : /* Block Text Paragraph Tabs */
        break;
    case 0x63 : /* Block Text Paragraph Bullets */
        break;
    case 0x64 : /* Block Text Paragraph Bullet Level */
        break;
    case 0x65 : /* Block Text Paragraph Line Horizontal Alignment */
        break;
    case 0x66 : /* Block Text Paragraph Line Vertical Alignment */
        break;
    case 0x67 : /* Block Text Paragragh Line Spacing */
        break;
    case 0x68 : /* Block Text Paragraph Word Wrap */
        break;
    case 0x70 : /* Block Text Forward Advance Distance */
        break;
    case 0x71 : /* Word Spacing */
        break;
    case 0x72 : /* External Leading */
        break;
    case 0x7a : /* set Gradient Offset */
        {
            long nHorzOffset = r_get_IX(r);
            long nVertOffset = r_get_IX(r);
            //(void)ImplGetUI16(); // nType
            c->vt->SetGradientOffset(c->userdata, nHorzOffset, nVertOffset );
            c->mnAct4PostReset |= ACT4_GRADIENT_ACTION;
        }
        break;
    case 0x7b : /* set Gradient Edge */
        c->mnAct4PostReset |= ACT4_GRADIENT_ACTION;
        break;
    case 0x7c : /* set Gradient Angle */
        c->vt->SetGradientAngle(c->userdata, r_get_I(r));
        c->mnAct4PostReset |= ACT4_GRADIENT_ACTION;
        break;
    case 0x7d : /* set Gradient Description */
        {
#if 0
            r_get_index(r); // -Wall is this needed?
            uint32_t  nNumberOfStages = r_get_index(r);
            uint32_t  i, nColorFrom = 0;
            uint32_t  nColorTo = 0xffffff;

            const size_t nRemainingSize = cgm_read_bytes(r);
            const size_t nMaxPossibleRecords = nRemainingSize/r->real_size;

            if (nNumberOfStages > nMaxPossibleRecords)
            {
                return -1;
            }

            for ( i = 0; i < nNumberOfStages; i++ )
            {
                r_get_real(r);
            }

            for ( i = 0; i <= nNumberOfStages; i++ )
            {
                //uint32_t nPara = mnParaSize + 24;
                if ( i == 0 )
                {
                    nColorTo = r_get_color(r, 0);
                    nColorFrom = nColorTo ^ 0xffffff;
                }
                else if ( i == 1 )
                    nColorFrom = r_get_color(r, 0);
                //mnParaSize = nPara;
            }
            if ( nNumberOfStages > 1 )
                c->vt->SetGradientStyle(c->userdata, 0xff );

            c->vt->SetGradientDescriptor(c->userdata, nColorFrom, nColorTo );
            c->mnAct4PostReset |= ACT4_GRADIENT_ACTION;
#endif
        }
        break;
    case 0x7e : /* set Gradient Style */
        {
#if 1
            //gradient_style_t nStyle = r_get_gradient_style(r);
#else
            uint32_t nStyle = ImplGetUI16();
#endif
            (void)r_get_R(r); // fRatio
            c->vt->SetGradientStyle(c->userdata, 0 /* nStyle */ );
            c->mnAct4PostReset |= ACT4_GRADIENT_ACTION;
        }
        break;
    case 0xff : /* inquire Font metrics */
        break;
    case 0xfe : /* inquire character widths */
        break;
    case 0xfd : /* set Text Font */
        break;
    case 0xfc : /* set current position */
        break;
    case 0xfb : /* set current position mode */
        break;
    case 0xfa : /* set character height mode */
        break;
    case 0xf9 : /* set Transform matrix 2D */
        break;
    case 0xf8 : /* set Transform matrix 3D */
        break;
    case 0xf7 : /* pop transformation state */
        break;
    case 0xf6 : /* clear transformation state */
        break;
    case 0xf5 : /* set character widths */
        break;
    case 0xf4 : /* set color name - for Pantone support */
        break;
    default:
        break;
    }
    return 0;
};
