/*
 * Copyright (C) 2000 – 2020  LibreOffice contributors
 * Copyright (C) 2021         Magnus Bergman <magnus@snisurset.net>
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * This file incorporates work covered by the following license notice:
 *
 *   Licensed to the Apache Software Foundation (ASF) under one or more
 *   contributor license agreements. See the NOTICE file distributed
 *   with this work for additional information regarding copyright
 *   ownership. The ASF licenses this file to you under the Apache
 *   License, Version 2.0 (the "License"); you may not use this file
 *   except in compliance with the License. You may obtain a copy of
 *   the License at http://www.apache.org/licenses/LICENSE-2.0 .
 */
#ifndef elements_h
#define elements_h

#include <loa-array.h>
#include <stdbool.h>

#include "cgmtypes.h"


typedef struct{
    int idx;
    color_t color;
}Bundle;

typedef struct{
    Bundle base;
    line_type_t type;
    double width;
}LineBundle;

typedef struct{
    Bundle base;
    MarkerType          eMarkerType;
    double              nMarkerSize;
}MarkerBundle;

typedef struct{
    Bundle base;
    int                 nTextFontIndex;
    TextPrecision       eTextPrecision;
    double              nCharacterExpansion;
    double              nCharacterSpacing;
}TextBundle;

typedef struct{
    Bundle base;
    interior_style_t   eFillInteriorStyle;
    int                nFillPatternIndex;
    int                nFillHatchIndex;
}FillBundle;

typedef struct{
    char *font_name;
    char *char_set_value;
    int type; /* bit 0 = 1 -> Italic, bit 1 = 1 -> Bold */
}FontEntry;

#define font_entry_init(e) memset(e, 0, sizeof(FontEntry))
void font_entry_done(FontEntry*);


typedef struct{
    uint32_t      nFontNameCount;
    uint32_t      nCharSetCount;
    //std::vector<FontEntry> aFontEntryList;
    loa_array_t font_entry_list;
}CGMFList;

CGMFList *cgm_font_list_init(CGMFList*);
void cgm_font_list_done(CGMFList*);
FontEntry*      CGMFList_GetFontEntry(CGMFList*, uint32_t );
void            CGMFList_InsertName(CGMFList*, const char *pSource, uint32_t nSize );
void            CGMFList_InsertCharSet(CGMFList*, const char *pSource, uint32_t nSize );


#define ASF_LINETYPE            (1<<0)
#define ASF_LINEWIDTH           (1<<1)
#define ASF_LINECOLOR           (1<<2)
#define ASF_MARKERTYPE          (1<<3)
#define ASF_MARKERSIZE          (1<<4)
#define ASF_MARKERCOLOR         (1<<5)    // NS
#define ASF_FILLINTERIORSTYLE   (1<<6)
#define ASF_HATCHINDEX          (1<<7)
#define ASF_PATTERNINDEX        (1<<8)
#define ASF_BITMAPINDEX         (1<<9)    // NS
#define ASF_FILLCOLOR           (1<<10)
#define ASF_EDGETYPE            (1<<11)
#define ASF_EDGEWIDTH           (1<<12)
#define ASF_EDGECOLOR           (1<<13)
#define ASF_TEXTFONTINDEX       (1<<14)
#define ASF_TEXTPRECISION       (1<<15)
#define ASF_CHARACTEREXPANSION  (1<<16)
#define ASF_CHARACTERSPACING    (1<<17)
#define ASF_TEXTCOLOR           (1<<18)

#define nBackGroundColor    aColorTable[ 0 ]


typedef struct{
    // ASPECT SOURCE FLAGS
    uint32_t            nAspectSourceFlags; // bit = 0 -> INDIVIDUAL
                                            //       1 -> BUNDLED

    /* Attributes */

    loa_array_t         line_list;
    int                 line_bundled;
    SpecMode            eLineWidthSpecMode;
    LineCapType         eLineCapType;
    LineJoinType        eLineJoinType;

    loa_array_t         marker_list;
    int                 marker_bundled;
    SpecMode            eMarkerSizeSpecMode;

    loa_array_t         edge_list;
    int                 edge_bundled;
    EdgeVisibility      eEdgeVisibility;
    SpecMode            eEdgeWidthSpecMode;

    loa_array_t         text_list;
    int                 text_bundled;
    double              nCharacterHeight;
    double              nCharacterOrientation[ 4 ];
    UnderlineMode       eUnderlineMode;
    uint32_t            nUnderlineColor;
    TextPath            eTextPath;
    TextAlignmentH      eTextAlignmentH;
    TextAlignmentV      eTextAlignmentV;
    double              nTextAlignmentHCont;
    double              nTextAlignmentVCont;
    long                nCharacterSetIndex;
    long                nAlternateCharacterSetIndex;
    CharacterCodingA    eCharacterCodingA;
    CGMFList            aFontList;

    loa_array_t         fill_list;
    int                 fill_bundled;
    loa_point_t         aFillRefPoint;
    //::std::map<uint32_t, HatchEntry> maHatchMap;

    /* Pattern attributes */

    /* Output control */

    /* Pick identifier */

    /* Symbol attributes */

}cgm_attributes_t;

cgm_attributes_t *cgm_attributes_init(cgm_attributes_t*);
void cgm_attributes_done(cgm_attributes_t*);

#define select_line_bundle(a,m) ((LineBundle*)loa_array_get(&(a)->line_list,(a)->nAspectSourceFlags&(m)?(a)->line_bundled:0))
#define select_marker_bundle(a,m) ((MarkerBundle*)loa_array_get(&(a)->marker_list,(a)->nAspectSourceFlags&(m)?(a)->marker_bundled:0))
#define select_edge_bundle(a,m) ((LineBundle*)loa_array_get(&(a)->edge_list,(a)->nAspectSourceFlags&(m)?(a)->edge_bundled:0))
#define select_text_bundle(a,m) ((TextBundle*)loa_array_get(&(a)->text_list,(a)->nAspectSourceFlags&(m)?(a)->text_bundled:0))
#define select_fill_bundle(a,m) ((FillBundle*)loa_array_get(&(a)->fill_list,(a)->nAspectSourceFlags&(m)?(a)->fill_bundled:0))

#define a_get_line_type(a) (select_line_bundle(a,ASF_LINETYPE)->type)
#define a_get_line_width(a) (select_line_bundle(a,ASF_LINEWIDTH)->width)
#define a_get_line_color(a) (select_line_bundle(a,ASF_LINECOLOR)->base.color)

#define a_get_edge_type(a) (select_edge_bundle(a,ASF_EDGETYPE)->type)
#define a_get_edge_width(a) (select_edge_bundle(a,ASF_EDGEWIDTH)->width)
#define a_get_edge_color(a) (select_edge_bundle(a,ASF_EDGECOLOR)->base.color)

#define a_get_text_color(a) (select_text_bundle(a,ASF_TEXTCOLOR)->base.color)
#define a_get_font_entry(a) CGMFList_GetFontEntry(&(a)->aFontList, (a)->nCharacterSetIndex)

#define a_get_fill_color(a) (select_fill_bundle(a,ASF_FILLCOLOR)->base.color)

typedef struct _CGMElements CGMElements;

/**
  This the set of attributes which will be saved using ...
*/
struct _CGMElements
{
    //long                nMetaFileVersion;
#if 0
    int                 nIntegerPrecision;  // maybe 1, 2, 4 Bytes
    int                 nIndexPrecision;    //   "      "       "
    RealPrecision       eRealPrecision;
    int                 nRealSize;          // maybe 4 or 8 bytes
    int                 nColorPrecision;    //   "      "       "
    int                 nColorIndexPrecision;//  "      "       "
#endif

    ScalingMode         eScalingMode;
    double              nScalingFactor;

#if 0
    VDCType             eVDCType;           // Integer / Real
    int                 nVDCIntegerPrecision;
    RealPrecision       eVDCRealPrecision;
    int                 nVDCRealSize;
#endif
    rect_t              aVDCExtent;
    rect_t              aVDCExtentMaximum;

#if 0
    DeviceViewPortMode  eDeviceViewPortMode;
#endif
    double              nDeviceViewPortScale;
    DeviceViewPortMap   eDeviceViewPortMap;
    DeviceViewPortMapH  eDeviceViewPortMapH;
    DeviceViewPortMapV  eDeviceViewPortMapV;
    rect_t              aDeviceViewPort;

    double              nMitreLimit;

    ClipIndicator       eClipIndicator;
    rect_t              aClipRect;

    ColorSelectionMode  eColorSelectionMode;
    uint32_t            nColorMaximumIndex;             // default 63
    uint32_t            nLatestColorMaximumIndex;       // default 63
    int8_t              aColorTableEntryIs[ 256 ];
    uint32_t            aColorTable[ 256 ];
    uint32_t            aLatestColorTable[ 256 ];
    uint32_t            nColorValueExtent[ 8 ]; // RGB, CMYK

    cgm_attributes_t    attr;

    Transparency        eTransparency;

    uint32_t            nAuxiliaryColor;

    // Delimiter Counts -> which will be increased by each 'begin' operation
    //                     and decreased by each 'end' operation
    bool                bSegmentCount;

    /* Control elements */

};

CGMElements *cgm_element_init(CGMElements*);
void cgm_element_done(CGMElements*);


void         CGMElements_ImplInsertHatch(CGMElements*, int32_t Key, int Style, long Distance, long Angle );

Bundle* cgm_get_bundle(loa_array_t*, int idx);
void         CGMElements_CopyAllBundles( const loa_array_t *Source, loa_array_t *Dest );

#endif /* elements_t */
