/* dmc_unrar - A dependency-free, single-file FLOSS unrar library
 *
 * Copyright (c) 2017, 2019-2020 by Sven Hesse (DrMcCoy) <drmccoy@drmccoy.de>
 *
 * dmc_unrar is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of
 * the License, or (at your option) any later version.
 *
 * dmc_unrar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * For the full text of the GNU General Public License version 2,
 * see <https://www.gnu.org/licenses/gpl-2.0.html>
 *
 * This code is based heavily on other FLOSS code. Please see the
 * bottom of this file for attribution and license information.
 */
#ifndef dmc_unrar_h
#define dmc_unrar_h

#include <stdbool.h>
#include <stddef.h>                     /* size_t */
#include <stdint.h>
#include <stdio.h>
#include <sys/types.h>                  /* off_t */

/* --- API types and macros --- */

/** The return code of a dmc_unrar operation. See dmc_unrar_strerror(). */
typedef enum {
	DMC_UNRAR_OK = 0,

	DMC_UNRAR_NO_ALLOC,
	DMC_UNRAR_ALLOC_FAIL,

	DMC_UNRAR_OPEN_FAIL,
	DMC_UNRAR_READ_FAIL,
	DMC_UNRAR_WRITE_FAIL,
	DMC_UNRAR_SEEK_FAIL,

	DMC_UNRAR_INVALID_DATA,

	DMC_UNRAR_ARCHIVE_EMPTY,

	DMC_UNRAR_ARCHIVE_IS_NULL,
	DMC_UNRAR_ARCHIVE_NOT_CLEARED,
	DMC_UNRAR_ARCHIVE_MISSING_FIELDS,

	DMC_UNRAR_ARCHIVE_NOT_RAR,
	DMC_UNRAR_ARCHIVE_UNSUPPORTED_ANCIENT,

	DMC_UNRAR_ARCHIVE_UNSUPPORTED_VOLUMES,
	DMC_UNRAR_ARCHIVE_UNSUPPORTED_ENCRYPTED,

	DMC_UNRAR_FILE_IS_INVALID,
	DMC_UNRAR_FILE_IS_DIRECTORY,

	DMC_UNRAR_FILE_SOLID_BROKEN,
	DMC_UNRAR_FILE_CRC32_FAIL,

	DMC_UNRAR_FILE_UNSUPPORTED_VERSION,
	DMC_UNRAR_FILE_UNSUPPORTED_METHOD,
	DMC_UNRAR_FILE_UNSUPPORTED_ENCRYPTED,
	DMC_UNRAR_FILE_UNSUPPORTED_SPLIT,
	DMC_UNRAR_FILE_UNSUPPORTED_LINK,
	DMC_UNRAR_FILE_UNSUPPORTED_LARGE,

	DMC_UNRAR_HUFF_RESERVED_SYMBOL,
	DMC_UNRAR_HUFF_PREFIX_PRESENT,
	DMC_UNRAR_HUFF_INVALID_CODE,

	DMC_UNRAR_PPMD_INVALID_MAXORDER,

	DMC_UNRAR_FILTERS_UNKNOWN,
	DMC_UNRAR_FILTERS_INVALID_FILTER_INDEX,
	DMC_UNRAR_FILTERS_REUSE_LENGTH_NEW_FILTER,
	DMC_UNRAR_FILTERS_INVALID_LENGTH,
	DMC_UNRAR_FILTERS_INVALID_FILE_POSITION,
	DMC_UNRAR_FILTERS_XOR_SUM_NO_MATCH,

	DMC_UNRAR_15_INVALID_FLAG_INDEX,
	DMC_UNRAR_15_INVALID_LONG_MATCH_OFFSET_INDEX,

	DMC_UNRAR_20_INVALID_LENGTH_TABLE_DATA,

	DMC_UNRAR_30_DISABLED_FEATURE_PPMD,
	DMC_UNRAR_30_DISABLED_FEATURE_FILTERS,

	DMC_UNRAR_30_INVALID_LENGTH_TABLE_DATA,

	DMC_UNRAR_50_DISABLED_FEATURE_FILTERS,

	DMC_UNRAR_50_INVALID_LENGTH_TABLE_DATA,
	DMC_UNRAR_50_BLOCK_CHECKSUM_NO_MATCH

} dmc_unrar_return;

/** The operating system a file was packed into a RAR. */
typedef enum {
	DMC_UNRAR_HOSTOS_DOS   = 0, /**< DOS, MS-DOS. */
	DMC_UNRAR_HOSTOS_OS2   = 1, /**< OS/2. */
	DMC_UNRAR_HOSTOS_WIN32 = 2, /**< Windows. */
	DMC_UNRAR_HOSTOS_UNIX  = 3, /**< Unix. */
	DMC_UNRAR_HOSTOS_MACOS = 4, /**< Mac OS. */
	DMC_UNRAR_HOSTOS_BEOS  = 5  /**< BeOS. */
} dmc_unrar_host_os;

typedef struct dmc_unrar_internal_state_tag dmc_unrar_internal_state;

/** A file entry within a RAR archive. */
typedef struct dmc_unrar_file_block_tag dmc_unrar_file_block;
struct dmc_unrar_file_block_tag {
	size_t index; /** The index of this file within the files array. */

	uint64_t start_pos; /**< The offset within the file *after* the whole file block header. */

	uint64_t flags; /**< flags describing the file. */

	uint16_t version; /**< RAR compression version for this file. */
	uint8_t  method;  /**< RAR compression method for this file. */

	uint64_t name_offset; /**< Offset to the name field. */
	uint64_t name_size;   /**< Size of the name field. */

	bool is_split;     /**< This file is a split file. */
	bool is_solid;     /**< This is a solid file. */
	bool is_link;      /**< This file is hard or symbolic link. */
	bool is_encrypted; /**< This file is encrypted. */

	uint64_t dict_size; /**< Dictionary size in bytes. */

	dmc_unrar_file_block *solid_start; /**< The first file entry in a solid block. */
	dmc_unrar_file_block *solid_prev;  /**< The previous file entry in a solid block. */
	dmc_unrar_file_block *solid_next;  /**< The next file entry in a solid block. */

    /* public part */

	uint64_t compressed_size;   /**< Size of the compressed file data, in bytes. */
	uint64_t uncompressed_size; /**< Size of the uncompressed file data, in bytes. */

	/** The operating system on which the file was packed into the RAR. */
	dmc_unrar_host_os host_os;

	bool has_crc; /**< Does this file entry have a checksum? */

	uint32_t crc;       /**< Checksum (CRC-32, 0xEDB88320 polynomial). */
	uint64_t unix_time; /**< File modification timestamp, POSIX epoch format. */

	/** File attributes, operating-system-specific.
	 *
	 *  The meaning depends on the host_os value:
	 *  - DMC_UNRAR_HOSTOS_DOS:   see dmc_unrar_windows_attribute
	 *  - DMC_UNRAR_HOSTOS_OS2:   ???
	 *  - DMC_UNRAR_HOSTOS_WIN32: see dmc_unrar_windows_attribute
	 *  - DMC_UNRAR_HOSTOS_UNIX:  see dmc_unrar_unix_attribute
	 *  - DMC_UNRAR_HOSTOS_MACOS: ???
	 *  - DMC_UNRAR_HOSTOS_BEOS:  ???
	 */
	uint64_t attrs;
};

typedef struct dmc_unrar_rar_context_tag dmc_unrar_rar_context;

typedef struct dmc_unrar_io_handler_tag {
	void *(*open)(const char *path);
	void (*close)(void *opaque);
	size_t (*read)(void *opaque, void *buffer, size_t n);
	bool (*seek)(void *opaque, off_t offset, int origin);
	off_t (*tell)(void *opaque);
} dmc_unrar_io_handler;

typedef struct dmc_unrar_io_tag {
	dmc_unrar_io_handler *funcs; /**< RAR file management functions. Must not be NULL. */
	void *opaque;                /**< Private data passed to funcs' pointers. */

	size_t size;       /**< Size of the IO stream. */

} dmc_unrar_io;

/** A RAR archive. */
typedef struct dmc_unrar_archive_tag {
	dmc_unrar_io io;

	/** Private internal state. */
	dmc_unrar_internal_state *internal_state;

} dmc_unrar_archive;

/** Initialize an IO structure. */
bool dmc_unrar_io_init(dmc_unrar_io *io, dmc_unrar_io_handler *handler, void *opaque);

#if DMC_UNRAR_DISABLE_STDIO != 1
/** Initialize an IO structure from a FILE*. */
bool dmc_unrar_io_init_from_file(dmc_unrar_io *io, FILE *file, bool *allocated_new_opaque);
#endif

/** Close an IO structure. */
void dmc_unrar_io_close(dmc_unrar_io *io);

/** Read from an IO structure. */
size_t dmc_unrar_io_read(dmc_unrar_io *io, void *buffer, size_t n);

/** Seek in an IO structure. */
bool dmc_unrar_io_seek(dmc_unrar_io *io, off_t offset, int origin);

/** Get the position in an IO structure. */
off_t dmc_unrar_io_tell(dmc_unrar_io *io);

/** Detect whether an IO structure contains a RAR archive. */
bool dmc_unrar_is_rar(dmc_unrar_io *io);

/** Detect whether the memory region contains a RAR archive. */
bool dmc_unrar_is_rar_mem(const void *mem, size_t size);

#if DMC_UNRAR_DISABLE_STDIO != 1
/* Detect whether this FILE contains a RAR archive. */
bool dmc_unrar_is_rar_file(FILE *file);
#endif /* DMC_UNRAR_DISABLE_STDIO */

/** Detect whether the file at this path contains a RAR archive.
 *
 *  Please note that on Windows, full UTF-8 paths only work when using the WIN32 API
 *  (see DMC_UNRAR_DISABLE_WIN32 above). Without the WIN32 API, only plain ASCII paths
 *  are supported on Windows.
 *
 *  @param  path The path of the file to dmc_unrar_io_default_handler and read out of.
 *               This must be UTF-8.
 *  @return true on success
 */
bool dmc_unrar_is_rar_path(const char *path);

/** Initialize/clear this archive struct.
 *
 *  @param  archive A valid pointer to an archive structure to initialize.
 *  @return DMC_UNRAR_OK on success. Any other value is an error condition.
 */
dmc_unrar_return dmc_unrar_archive_init(dmc_unrar_archive *archive);

/** Open this RAR archive, reading its block and file headers.
 *  The io field must be initialized.
 *  The func_alloc, func_realloc, func_free and opaque_mem fields may be set.
 *  All other fields must have been cleared.
 *
 *  @param  archive Pointer to the archive structure to use. Needs to be a valid
 *                  pointer, with the fields properly initialized and set.
 *  @return DMC_UNRAR_OK if the archive was successfully opened. Any other value
 *          describes an error condition.
 */
dmc_unrar_return dmc_unrar_archive_open(dmc_unrar_archive *archive);

/** Open this RAR archive from a memory block, reading its block and file headers.
 *  The func_alloc, func_realloc, func_free and opaque_mem fields may be set.
 *  All other fields must have been cleared.
 *
 *  @param  archive Pointer to the archive structure to use. Needs to be a valid
 *                  pointer, with the fields properly initialized and set.
 *  @param  mem Pointer to a block of memory to read the RAR file out of.
 *  @param  size Size of the RAR memory region.
 *  @return DMC_UNRAR_OK if the archive was successfully opened. Any other value
 *          describes an error condition.
 */
dmc_unrar_return dmc_unrar_archive_open_mem(dmc_unrar_archive *archive,
	const void *mem, size_t size);

#if DMC_UNRAR_DISABLE_STDIO != 1
/** Open this RAR archive from a stdio FILE, reading its block and file headers.
 *  The func_alloc, func_realloc, func_free and opaque_mem fields may be set.
 *  All other fields must have been cleared.
 *
 *  The stdio FILE will be taken over and will be closed when the archive is
 *  closed with dmc_unrar_archive_close().
 *
 *  @param  archive Pointer to the archive structure to use. Needs to be a valid
 *                  pointer, with the fields properly initialized and set.
 *  @param  file The stdio FILE structure to read out of.
 *  @return DMC_UNRAR_OK if the archive was successfully opened. Any other value
 *          describes an error condition.
 */
dmc_unrar_return dmc_unrar_archive_open_file(dmc_unrar_archive *archive, FILE *file);
#endif /* DMC_UNRAR_DISABLE_STDIO */

/** Open this RAR archive from a path, opening the file with dmc_unrar_io_default_handler,
 *  and reading its block and file headers. The func_alloc, func_realloc, func_free and
 *  opaque_mem fields may be set. All other fields must have been cleared.
 *
 *  Please note that on Windows, full UTF-8 paths only work when using the WIN32 API
 *  (see DMC_UNRAR_DISABLE_WIN32 above). Without the WIN32 API, only plain ASCII paths
 *  are supported on Windows.
 *
 *  @param  archive Pointer to the archive structure to use. Needs to be a valid
 *                  pointer, with the fields properly initialized and set.
 *  @param  path The path of the file to dmc_unrar_io_default_handler and read out of.
 *               This must be UTF-8.
 *  @return DMC_UNRAR_OK if the archive was successfully opened. Any other value
 *          describes an error condition.
 */
dmc_unrar_return dmc_unrar_archive_open_path(dmc_unrar_archive *archive, const char *path);

/** Close this RAR archive again.
 *
 *  All allocated memory will be freed. */
void dmc_unrar_archive_close(dmc_unrar_archive *archive);

/** Return the number of file entries in this RAR archive. */
size_t dmc_unrar_get_file_count(dmc_unrar_archive *archive);

dmc_unrar_file_block *dmc_unrar_get_file(dmc_unrar_archive *archive, size_t index);

/** Get the filename of a RAR file entry, UTF-8 encoded and \0-terminated.
 *
 *  Note: the filename is *not* checked to make sure it contains fully
 *  valid UTF-8 data. Use dmc_unrar_unicode_is_valid_utf8() and/or
 *  dmc_unrar_unicode_make_valid_utf8() for that.
 *
 *  Returns the number of bytes written to filename. If filename is NULL, this function
 *  returns the number of bytes needed to fully store the filename.
 */
size_t dmc_unrar_get_filename(dmc_unrar_archive *archive, dmc_unrar_file_block *file,
	char *filename, size_t filename_size);

/** Is this file entry a directory? */
bool dmc_unrar_file_is_directory(dmc_unrar_archive *archive, dmc_unrar_file_block *file);

/** Check if we support extracted this file entry.
 *
 *  If we do support extracting this file entry, DMC_UNRAR_OK is returned.
 *  Otherwise, the return code gives an idea why we don't have support. */
dmc_unrar_return dmc_unrar_file_is_supported(dmc_unrar_archive *archive, dmc_unrar_file_block *file);

/** Extract a file entry into a pre-allocated memory buffer.
 *
 *  @param  archive The archive to extract from.
 *  @param  index The index of the file entry to extract.
 *  @param  buffer The pre-allocated memory buffer to extract into.
 *  @param  buffer_size The size of the pre-allocated memory buffer.
 *  @param  uncompressed_size If != NULL, the number of bytes written
 *          to the buffer will be stored here.
 *  @param  validate_crc If true, validate the uncompressed data against
 *          the CRC-32 stored within the archive. If the validation fails,
 *          this counts as an error (DMC_UNRAR_FILE_CRC32_FAIL).
 *  @return An error condition, or DMC_UNRAR_OK if extraction succeeded.
 */
dmc_unrar_return dmc_unrar_extract_file_to_mem(dmc_unrar_archive *archive, dmc_unrar_file_block *file,
	void *buffer, size_t buffer_size, size_t *uncompressed_size, bool validate_crc);

/** The callback function for dmc_unrar_extract_file_with_callback().
 *
 *  Note that even with small buffer slices, decompressing a buffer
 *  full might take an unexpected long time, if the requested file
 *  is part of a solid block and/or uses the PPMd decoder.
 *
 *  @param  opaque Opaque memory pointer for personal use.
 *  @param  buffer Pointer to the buffer where the current part of the
 *          extracted file resides. Can be changed, to use a different
 *          buffer for further extraction. Can be set to NULL to let
 *          dmc_unrar_extract_file_with_callback() allocate its own
 *          internal buffer.
 *  @param  buffer_size Size of the buffer. Can be modified, to use
 *          a different buffer size for further extraction.
 *  @param  uncompressed_size Number of bytes of extracted file waiting
 *          in the buffer.
 *  @param  err In combination with returning false, the callback can
 *          set this parameter to something other than DMC_UNRAR_OK to
 *          signal an error. dmc_unrar_extract_file_with_callback() will
 *          return with that error condition.
 *  @return true if extraction should continue, false otherwise.
 */
typedef bool (*dmc_unrar_extract_callback_func)(void *opaque, void **buffer,
	size_t *buffer_size, size_t uncompressed_size, dmc_unrar_return *err);

/** Extract a file entry using a callback function.
 *
 *  Extract into the buffer of buffer_size, calling callback every time the
 *  buffer has been filled (or all the input has been processed).
 *
 *  @param  archive The archive to extract from.
 *  @param  index The index of the file entry to extract.
 *  @param  buffer The pre-allocated memory buffer to extract into. Can be
 *          NULL to mean that a buffer of buffer_size should be allocated.
 *  @param  buffer_size The size of the output buffer.
 *  @param  uncompressed_size If != NULL, the total number of bytes written
 *          to the buffer will be stored here.
 *  @param  validate_crc If true, validate the uncompressed data against
 *          the CRC-32 stored within the archive. If the validation fails,
 *          this counts as an error (DMC_UNRAR_FILE_CRC32_FAIL).
 *  @param  opaque Opaque memory pointer to pass to the callback.
 *  @param  callback The callback to call.
 *  @return An error condition, or DMC_UNRAR_OK if extraction succeeded.
 */
dmc_unrar_return dmc_unrar_extract_file_with_callback(dmc_unrar_archive *archive,
	dmc_unrar_file_block *file, void *buffer, size_t buffer_size,
	size_t *uncompressed_size, bool validate_crc, void *opaque,
	dmc_unrar_extract_callback_func callback);

typedef enum {
	DMC_UNRAR_GENERATION_INVALID = 0,

	DMC_UNRAR_GENERATION_ANCIENT,
	DMC_UNRAR_GENERATION_RAR4,
	DMC_UNRAR_GENERATION_RAR5
} dmc_unrar_generation;

typedef struct dmc_unrar_block_header_tag {
	uint64_t start_pos; /**< The offset within the file the block start at. */
	uint64_t extra_pos; /**< The offset within the file the extra header data starts. */

	uint64_t type; /**< The type of the block. */

	uint32_t crc;   /**< Checksum. */
	uint64_t flags; /**< flags describing this block. */

	uint64_t header_size; /**< Size of the full block header in bytes. */
	uint64_t data_size;   /**< Size of the extra block data in bytes. */

	uint64_t extra_size; /** Size of extra file properties in RAR5, in bytes. */

} dmc_unrar_block_header;

struct dmc_unrar_internal_state_tag {
	/** RAR generation. RAR4 (1.5 - 3.6) vs RAR5 (5.0). */
	dmc_unrar_generation generation;

	uint16_t archive_flags; /**< Global archive flags. */

	dmc_unrar_block_header *comment; /**< Archive comments block. */

	size_t block_count;    /**< Number of blocks in this RAR archive. */
	dmc_unrar_block_header *blocks;  /**< All blocks in this RAR archive. */
	size_t block_capacity; /**< Memory capacity of the blocks array. */

	size_t file_count;    /**< Number of files (and directories) in this RAR archive. */
	dmc_unrar_file_block *files;    /**< All files (and directories) in this RAR archive. */
	size_t file_capacity; /**< Memory capacity of the files array. */

	/** Saved unpack context, for sequential solid block unpacking. */
	dmc_unrar_rar_context *unpack_context;
};

#endif /* dmc_unrar_h */
