/*
 * Copyright (C) 2019 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <libdjvu/ddjvuapi.h>
#include <stdlib.h>                     /* malloc(), free() */


#define LOAD_ALL_PAGES


typedef struct{
#ifdef LOAD_ALL_PAGES
    cairo_surface_t *surface;
#endif
    int width;
    int height;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    ddjvu_context_t *ctx;
    ddjvu_format_t *format;
    ddjvu_document_t *doc;
    page_t *page;
};


static const char **
_djvu_get_mime_types()
{
    static const char *mime_types[] = {
        "image/vnd.djvu",
        "image/vnd.djvu+multipage",
        "image/x-djvu",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_djvu_new(const char *mime_type, abydos_plugin_info_t *info)
{
    static unsigned int masks[4] = {0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000};
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;

    h->ctx = ddjvu_context_create(PACKAGE);
    h->format = ddjvu_format_create(DDJVU_FORMAT_RGBMASK32, 4, masks);
    ddjvu_format_set_row_order(h->format, 1);
    h->page = NULL;

    return h;
}


static void
_djvu_free(abydos_plugin_handle_t *h)
{
    ddjvu_context_release(h->ctx);
    ddjvu_format_release(h->format);
    if(h->doc)
        ddjvu_document_release(h->doc);
    if(h->page){
        int i;
        for(i=0;i<h->info->page_count;++i)
            if(h->page[i].surface)
                cairo_surface_destroy(h->page[i].surface);
        free(h->page);
    }
    free(h);
}


static void
_handle_events(abydos_plugin_handle_t *h, int wait)
{
    const ddjvu_message_t *msg;
    if(wait)
        ddjvu_message_wait(h->ctx);
    while((msg = ddjvu_message_peek(h->ctx))){
        DEBUG_p("got message");
        ddjvu_message_pop(h->ctx);
    }
}


static cairo_surface_t *
_surface_from_page(ddjvu_page_t *page, ddjvu_format_t *format)
{
    ddjvu_rect_t rect;
    cairo_surface_t *surface;
    unsigned char *dst;
    int stride;
    rect.x = 0;
    rect.y = 0;
    rect.w = ddjvu_page_get_width(page);
    rect.h = ddjvu_page_get_height(page);
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, rect.w, rect.h);
    dst = cairo_image_surface_get_data(surface);
    stride = cairo_image_surface_get_stride(surface);
    ddjvu_page_render(page,DDJVU_RENDER_COLOR,&rect,&rect,format,stride,(char*)dst);
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_djvu_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    int i;
    h->doc = ddjvu_document_create_by_filename(h->ctx, filename, 1);
    if(!h->doc)
        return -1;
    _handle_events(h, 1);
    h->info->page_count = ddjvu_document_get_pagenum(h->doc);
    DEBUG_p("pages: %d",h->info->page_count);

    h->page = (page_t*)malloc(h->info->page_count * sizeof(page_t));
    for(i=0;i<h->info->page_count;++i){
        ddjvu_page_t *page = ddjvu_page_create_by_pageno(h->doc, i);
        while(!ddjvu_page_decoding_done(page))
            _handle_events(h, 1);
#ifdef LOAD_ALL_PAGES
        h->page[i].surface = _surface_from_page(page, h->format);
#endif
        h->page[i].width = ddjvu_page_get_width(page);
        h->page[i].height = ddjvu_page_get_height(page);
        DEBUG_p("dimensions: %dx%d", h->page[i].width, h->page[i].height);
        ddjvu_page_release(page);
    }

    return 0;
}


static cairo_surface_t *
_djvu_get_image_surface(abydos_plugin_handle_t *h, int pageno, cairo_bool_t *visible_layers,int prefered_variant, double time, int frame)
{
#ifdef LOAD_ALL_PAGES
    return cairo_surface_reference(h->page[pageno].surface);
#else
    ddjvu_page_t *page = ddjvu_page_create_by_pageno(h->doc, pageno);
    cairo_surface_t *surface;
    while(!ddjvu_page_decoding_done(page))
        _handle_events(h, 1);
    surface = _surface_from_page(page,h->format);
    ddjvu_page_release(page);
    return surface;
#endif
}


static void
_djvu_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _djvu_get_mime_types,
        .new = _djvu_new,
        .free = _djvu_free,
        .create_from_file = _djvu_create_from_file,
        .get_image_surface = _djvu_get_image_surface,
        .get_page_info = _djvu_get_page_info
    };
    return &p;
}
