/*
 * Copyright (C) 2021 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <nil-iff.h>
#include <nil-mcga.h>
#include <stdlib.h>
#include <string.h>


#define MLDF_MLDF IFF_ID('M','L','D','F')
#define MLDF_BMHD IFF_ID('B','M','H','D')
#define MLDF_GRAB IFF_ID('G','R','A','B')
#define MLDF_CMAP IFF_ID('C','M','A','P')
#define MLDF_DPPV IFF_ID('D','P','P','V')
#define MLDF_CRNG IFF_ID('C','R','N','G')
#define MLDF_BODY IFF_ID('B','O','D','Y')

#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_mldf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-mldf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_mldf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_mldf_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


typedef struct{
    nile_t ni;
}parse_context_t;

/*

BMHD:
---cabin.mld----------------------------------------------------------------
__width_____    __height____
0x01    0x40    0x00    0xc8    0x00    0x00    0x00    0x00

0x08    0x00    0xff    0x00    0x00    0x00    0x05    0x06
__screen_w__    __screen_h__
0x01	0x40	0x00	0xc8
---cdoor.bru----------------------------------------------------------------

0x00	0x39	0x00	0x19	0x00	0x00	0x00	0x00
        _!!_
0x08	0x02	0xff	0x00	0x00	0x00	0x05	0x06

0x01	0x40	0x00	0xc8


*/

static int
_on_chunk(void *userdata, uint32_t id, const char *data, size_t len)
{
    parse_context_t *pc = userdata;
    DEBUG_p("id: %.4s, size: %d",(char*)&id, (int)len);
    switch(id){
    case MLDF_BMHD:
        if(len < 4)
            return -1;
        pc->ni.type = NILE_TYPE_INDEXED;
        pc->ni.width = get16(data);
        pc->ni.height = get16(data+2);
        pc->ni.stride = ((pc->ni.width-1)/16+1)*16;
        pc->ni.colors = 256;
        pc->ni.bpc = 3;
        pc->ni.format = NILE_FORMAT_r8g8b8;
        break;
    case MLDF_GRAB:
        break;
    case MLDF_CMAP:
        if(len < 768)
            return -1;
        pc->ni.palette = (uint8_t*)data;
        break;
    case MLDF_DPPV:
        break;
    case MLDF_CRNG:
        break;
    case MLDF_BODY:
        DEBUG_p("expected size: %d", pc->ni.stride * pc->ni.height);
        if(len < pc->ni.stride * pc->ni.height)
            return -1;
        pc->ni.pixels = (uint8_t*)data;
        break;
    default:
        DEBUG_p("unknwon chunk");
        break;
    }
    return 0;
}


static int
_mldf_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    parse_context_t pc;
    int res;

    nile_init(&pc.ni);
    res = nil_iff_parse(MLDF_MLDF, -8, data, len, _on_chunk, &pc);
    if(!res){
        h->info->width = pc.ni.width;
        h->info->height = pc.ni.height;
        h->info->pixel_ratio = 5.0/6;
        if(!pc.ni.palette)
            pc.ni.palette = (uint8_t*)nil_mcga_palette;
        h->surface = nil_cairo_surface_create(&pc.ni, 0);
    }
    nile_done(&pc.ni);
    return res;
}


static cairo_surface_t *
_mldf_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _mldf_get_mime_types,
        .new = _mldf_new,
        .free = _mldf_free,
        .create_from_data = _mldf_create_from_data,
        .get_image_surface = _mldf_get_image_surface
    };
    return &p;
}
