/*
 * Copyright (C) 2023  Magnus Bergman <magnus@snisurset.net>
 * Copyright (c) 2021  Dominic Szablewski - https://phoboslab.org
 *
 * SPDX-License-Identifier: MIT
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#define QOI_OP_INDEX  0x00 /* 00xxxxxx */
#define QOI_OP_DIFF   0x40 /* 01xxxxxx */
#define QOI_OP_LUMA   0x80 /* 10xxxxxx */
#define QOI_OP_RUN    0xc0 /* 11xxxxxx */
#define QOI_OP_RGB    0xfe /* 11111110 */
#define QOI_OP_RGBA   0xff /* 11111111 */

#define QOI_MASK_2    0xc0 /* 11000000 */

#define QOI_COLOR_HASH(C) (C.r*3 + C.g*5 + C.b*7 + C.a*11)
#define QOI_MAGIC (('q') << 24 | ('o') << 16 | ('i') <<  8 | 'f')
#define QOI_HEADER_SIZE 14

#define QOI_PIXELS_MAX ((unsigned int)400000000)

typedef struct{
    uint8_t r;
    uint8_t g;
    uint8_t b;
    uint8_t a;
}rgba_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_qoi_get_mime_types()
{
    static const char *mime_types[] = {
        "image/qoi",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_qoi_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;

    return h;
}


static void
_qoi_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get32(a) be32toh(*(uint32_t*)(a))


static int
_qoi_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    uint8_t *src = (uint8_t*)data;
    uint8_t *end = src + len;
    unsigned int header_magic;
    uint8_t bpc;
    uint8_t colorspace;
    uint32_t *dst;
    int rowskip;
    rgba_t index[64];
    rgba_t px;
    int run = 0;
    int y;

    if(len < QOI_HEADER_SIZE)
        return -1;

    /* read header */

    header_magic = get32(src + 0);
    if(header_magic != QOI_MAGIC)
        return -1;
    h->info->width = get32(src + 4);
    h->info->height = get32(src + 8);
    if(h->info->width == 0 || h->info->height == 0 || h->info->height >= QOI_PIXELS_MAX / h->info->width)
        return -1;
    bpc = src[12];
    colorspace = src[13];
    if(colorspace > 1)
        return -1;
    src += QOI_HEADER_SIZE;

    /* create surface */

    switch(bpc){
    case 3:
        h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, h->info->width, h->info->height);
        break;
    case 4:
        h->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, h->info->width, h->info->height);
        break;
    default:
        return -1;
    }
    dst = (uint32_t*)cairo_image_surface_get_data(h->surface);
    if(!dst)
        return -1;
    rowskip = cairo_image_surface_get_stride(h->surface)/sizeof(uint32_t) - h->info->width;

    /* decode body */

    memset(index, 0, sizeof(index));
    px.r = 0;
    px.g = 0;
    px.b = 0;
    px.a = 255;

    for(y=0;y<h->info->height;++y){
        int x;
        for(x=0;x<h->info->width;++x){
            if(run > 0){
                run--;
            }else if(src < end){
                int b1 = *src++;
                if(QOI_OP_RGB == b1){
                    if(src + 3 > end)
                        return -1;
                    px.r = *src++;
                    px.g = *src++;
                    px.b = *src++;
                }else if(QOI_OP_RGBA == b1){
                    if(src + 4 > end)
                        return -1;
                    px.r = *src++;
                    px.g = *src++;
                    px.b = *src++;
                    px.a = *src++;
                }else if(QOI_OP_INDEX == (b1 & QOI_MASK_2)){
                    px = index[b1];
                }else if(QOI_OP_DIFF == (b1 & QOI_MASK_2)){
                    px.r += ((b1 >> 4) & 0x03) - 2;
                    px.g += ((b1 >> 2) & 0x03) - 2;
                    px.b += ( b1       & 0x03) - 2;
                }else if(QOI_OP_LUMA == (b1 & QOI_MASK_2)){
                    int b2;
                    int vg;
                    if(src + 1 > end)
                        return -1;
                    b2 = *src++;
                    vg = (b1 & 0x3f) - 32;
                    px.r += vg - 8 + ((b2 >> 4) & 0x0f);
                    px.g += vg;
                    px.b += vg - 8 +  (b2       & 0x0f);
                }else if(QOI_OP_RUN == (b1 & QOI_MASK_2)){
                    run = (b1 & 0x3f);
                }
                index[QOI_COLOR_HASH(px) % 64] = px;
            }
            if(4==bpc){
                int r = px.a * px.r / 255;
                int g = px.a * px.g / 255;
                int b = px.a * px.b / 255;
                *dst++ = r << 16 | g << 8 | b << 0 | px.a << 24;
            }else{
                *dst++ = px.r << 16 | px.g << 8 | px.b << 0;
            }
        }
        dst += rowskip;
    }
    return 0;
}


static cairo_surface_t *
_qoi_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _qoi_get_mime_types,
        .new = _qoi_new,
        .free = _qoi_free,
        .create_from_data = _qoi_create_from_data,
        .get_image_surface = _qoi_get_image_surface
    };
    return &p;
}
