/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <openjpeg.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#define ENABLE_JPX

struct _abydos_plugin_handle_t{
    OPJ_CODEC_FORMAT format;
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_openjp2_get_mime_types()
{
    static const char *mime_types[] = {
        "image/jp2",
        "image/jpeg2000",
#ifdef ENABLE_JPX
        "image/jpx",
#endif
        /*
        "image/jpm",
        "image/mj2",
        */
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_openjp2_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    if(!strcasecmp(mime_type,"image/jp2")|
       !strcasecmp(mime_type,"image/jpeg2000"))
    {
        h->format = OPJ_CODEC_JP2;
#ifdef ENABLE_JPX
    }else if(!strcasecmp(mime_type,"image/jpx")){
        h->format = OPJ_CODEC_JPX;
#endif
    /*
    }else if(!strcasecmp(mime_type,"image/jpm")){
    }else if(!strcasecmp(mime_type,"image/mj2")){
    */

    /*
    ".JP2" -> OPJ_CODEC_JP2
    ".J2K" -> OPJ_CODEC_J2K
    ".JPP" -> OPJ_CODEC_JPP
    ".JPT" -> OPJ_CODEC_JPT
    ".JPX" -> OPJ_CODEC_JPX
    */

    }else{
        free(h);
        return NULL;
    }
    h->info = info;
    h->info->threadsafe = 1;

    h->surface = NULL;
    return h;
}


static void
_openjp2_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static cairo_surface_t *
_surface_from_image(opj_image_t *img)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,img->comps[0].w,img->comps[0].h);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - img->comps[0].w;
    off_t offs = 0;
    int *rp;
    int *gp;
    int *bp;
    int *ap;
    int y;
    switch(img->numcomps){
    case 1: /* gray */
        gp = img->comps[0].data;
        for(y=0;y<img->comps[0].h;++y){
            int x;
            for(x=0;x<img->comps[0].w;++x){
                int g = *gp++;
                *dst++ = g<<16 | g<< 8 | g<< 0 | 255<<24;
            }
            dst += skip;
        }
        break;
        break;
    case 2: /* gray + alpha */
        gp = img->comps[0].data;
        ap = img->comps[1].data;
        for(y=0;y<img->comps[0].h;++y){
            int x;
            for(x=0;x<img->comps[0].w;++x){
                int alpha = ap[offs];
                int g = alpha*gp[offs]/255;
                *dst++ = g<<16 | g<< 8 | g<< 0 | alpha<<24;
                ++offs;
            }
            dst += skip;
        }
        break;
        break;
    case 3: /* RGB */
        rp = img->comps[0].data;
        gp = img->comps[1].data;
        bp = img->comps[2].data;
        for(y=0;y<img->comps[0].h;++y){
            int x;
            for(x=0;x<img->comps[0].w;++x){
                int r = *rp++;
                int g = *gp++;
                int b = *bp++;
                *dst++ = r<<16 | g<< 8 | b<< 0 | 255<<24;
            }
            dst += skip;
        }
        break;
        break;
    case 4: /* RGBA */
        rp = img->comps[0].data;
        gp = img->comps[1].data;
        bp = img->comps[2].data;
        ap = img->comps[3].data;
        for(y=0;y<img->comps[0].h;++y){
            int x;
            for(x=0;x<img->comps[0].w;++x){
                int alpha = ap[offs];
                int r = alpha*rp[offs]/255;
                int g = alpha*gp[offs]/255;
                int b = alpha*bp[offs]/255;
                *dst++ = r<<16 | g<< 8 | b<< 0 | alpha<<24;
                ++offs;
            }
            dst += skip;
        }
        break;
    default:
        cairo_surface_destroy(surface);
        return NULL;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_openjp2_create_from_stream(abydos_plugin_handle_t *h, opj_stream_t *stream)
{
    opj_codec_t *codec = opj_create_decompress(h->format);
    opj_dparameters_t params;
    opj_image_t *image = NULL;

    /* catch events using our callbacks and give a local context */
    /*
    opj_set_info_handler(l_codec, info_callback,00);
    opj_set_warning_handler(l_codec, warning_callback,00);
    opj_set_error_handler(l_codec, error_callback,00);
    */

    opj_set_default_decoder_parameters(&params);
    if(!opj_setup_decoder(codec, &params)){
        DEBUG_p("failed to setup the decoder");
        opj_destroy_codec(codec);
        return -1;
    }

    /* Read the main header of the codestream and if necessary the JP2 boxes*/
    if(! opj_read_header(stream, codec, &image)){
        DEBUG_p("failed to read the header");
        opj_destroy_codec(codec);
        opj_image_destroy(image);
        return -1;
    }

    if(!opj_get_decoded_tile(codec, stream, image, params.tile_index)) {
        DEBUG_p("failed to decode tile");
        opj_destroy_codec(codec);
        opj_image_destroy(image);
        return -1;
    }
    DEBUG_p("tile %d is decoded", params.tile_index);
    opj_destroy_codec(codec);

    /* FIXME!!
    if(image->color_space == OPJ_CLRSPC_SYCC)
        color_sycc_to_rgb(image);
    */

    /* FIXME!!
    if(image->icc_profile_buf){
        color_apply_icc_profile(image);
        free(image->icc_profile_buf);
        image->icc_profile_buf = NULL;
        image->icc_profile_len = 0;
    }
    */

    h->surface = _surface_from_image(image);
    opj_image_destroy(image);
    if(h->surface){
        h->info->width = cairo_image_surface_get_width(h->surface);
        h->info->height = cairo_image_surface_get_height(h->surface);
        return 0;
    }else{
        return -1;
    }
}


#if 0
static int
_openjp2_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    FILE *f;
    int res;
    f = fmemopen((void*)data,len,"r");
    if(!f)
        return -1;
    res = _openjp2_create_from_stream(h,f);
    fclose(f);
    return res;
}
#endif


static int
_openjp2_create_from_file(abydos_plugin_handle_t *h, const char *filename)
{
    opj_stream_t *stream = opj_stream_create_default_file_stream(filename,1);
    int res;
    if(!stream){
        DEBUG_p("failed to create the stream from the file");
        return -1;
    }
    res = _openjp2_create_from_stream(h,stream);
    opj_stream_destroy(stream);
    return res;
}


static cairo_surface_t *
_openjp2_get_image_surface(abydos_plugin_handle_t *h,
                        int page,
                        cairo_bool_t *layer_visibility,
                        int variant,
                        double time,
                        int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _openjp2_get_mime_types,
        .new = _openjp2_new,
        .free = _openjp2_free,
        //.create_from_data = _openjp2_create_from_data,
        .create_from_file = _openjp2_create_from_file,
        .get_image_surface = _openjp2_get_image_surface
    };
    return &p;
}
