/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <nil-cairo.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "nil-pc88.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_am_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-artmaster",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_am_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_am_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


#define get16(a) le16toh(*(uint16_t*)(a))


static int
_am_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    nile_t ni;
    const uint8_t *p;
    const uint8_t *end;
    const uint8_t *flag;
    int i;

    if(len < 16+6+18)
        goto error_truncated;
    if(memcmp(data, "SS_SIF    ",10)){
        h->info->error = "Invalid header";
        return -1;
    }

    p = (uint8_t*)data;
    end = p + len;

    p += 16;

    flag = p;
    p += 6;
    p += 18;

    nile_init(&ni);

    if('I' == flag[0]){
        /* Skip over something */
        int bytes;
        if(p+2 > end)
            goto error_truncated;
        bytes = get16(p);
        p += bytes;
    }

    ni.bpc = 2;
    ni.format = NILE_FORMAT_r5g6b5;
    if('H' == flag[1]){
        /* FIXME!! read HSL palette */
    }else{
        nil_pc88_default_palette(&ni);
    }

    if('B' == flag[2]){
        /* Skip over something */
        int bytes;
        if(p+2 > end)
            goto error_truncated;
        bytes = get16(p);
        p += bytes;
    }

    nile_ensure_bitplanes(&ni, 640, 200, 3, 1);
    for(i=0;i<3;++i){
        uint8_t *dst;
        int offs;
        int count = 0;
        int last = -1;
        switch(flag[3+i]){
        case 'B':
            dst = ni.bitplanes[0];
            break;
        case 'R':
            dst = ni.bitplanes[1];
            break;
        case 'G':
            dst = ni.bitplanes[2];
            break;
        default:
            goto error_invalid;
        }
        DEBUG_p("plane %d",i);
        for(offs=0;offs<16000;++offs){
            int b;
            if(count <= 0){
                if(p >= end)
                    goto error_truncated;
                b = *p++;
                if(b == last){
                    /* It's a runlength. */
                    if(p >= end)
                        goto error_truncated;
                    count = *p++;
                    if(--count < 0)
                        count = 255;
                    if(0 == count)
                        goto error_invalid;
                    last = -1;
                }else{
                    count = 1;
                    last = b;
                }
            }
            dst[offs] = b;
            --count;
        }
    }

    h->info->width = 640;
    h->info->height = 200;
    h->info->pixel_ratio = 5.0/12;
    h->surface = nil_cairo_surface_create(&ni, 0);
    nile_done(&ni);

    return 0;
error_invalid:
    nile_done(&ni);
    DEBUG_p("invalid data");
    h->info->error = "Invalid data";
    return -1;
error_truncated:
    nile_done(&ni);
    DEBUG_p("truncated data");
    h->info->error = "Truncated data";
    return -1;
}


static cairo_surface_t *
_am_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _am_get_mime_types,
        .new = _am_new,
        .free = _am_free,
        .create_from_data = _am_create_from_data,
        .get_image_surface = _am_get_image_surface
    };
    return &p;
}
