/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdlib.h>                     /* malloc(), free() */

#include "debug.h"
#include "teletext_cons.h"
#include "teletext_dec.h"


void teletext_reset(TeletextState* s){
    static const TeletextState default_state = {
        .bg = 0,
        .fg = 7,
        .flash = 0,
        .graphic = 0,
        .doubleheight = 0
    };
    *s = default_state;
};


int teletext_update(TeletextState* s,int ch){
    ch &= 0x7f;
    switch(ch){
    case 0x0a:
    case 0x0b:
    case 0x0e:
    case 0x0f:
    case 0x1b:
        /* nothing */
        break;
    case 0x00: /* Alpha Black */
    case 0x01: /* Alpha Red */
    case 0x02: /* Alpha Green */
    case 0x03: /* Alpha Yellow */
    case 0x04: /* Alpha Blue */
    case 0x05: /* Alpha Magenta */
    case 0x06: /* Alpha Cyan */
    case 0x07: /* Alpha White * */
        s->fg = ch;
        s->graphic = 0;
        break;
    case 0x08: /* Flash */
        s->flash = 1;
        break;
    case 0x09: /* Steady * */
        s->flash = 0;
        break;
    case 0x0c: /* Normal Height * */
        s->doubleheight = 1;
        break;
    case 0x0d: /* Double height */
        s->doubleheight = 0;
        break;
    case 0x10: /* Graphic Black */
    case 0x11: /* Graphic Red */
    case 0x12: /* Graphic Green */
    case 0x13: /* Graphic Yellow */
    case 0x14: /* Graphic Blue */
    case 0x15: /* Graphic Magenta */
    case 0x16: /* Graphic Cyan */
    case 0x17: /* Graphic White */
        s->fg = ch-0x10;
        s->graphic = 1;
        break;
    case 0x18: /* Conceal Display */
    case 0x19: /* Contiguous Graphics * */
    case 0x1a: /* Separated Graphics */
        /* FIXME!! */
        break;
    case 0x1c: /* Black Background * */
        s->bg = 0;
        break;
    case 0x1d: /* New Background */
        s->bg = s->fg;
        break;
    case 0x1e: /* Hold Graphics */
    case 0x1f: /* Release Graphics * */
        /* FIXME!! */
        break;
    default:
        return ch;
    }
    return ' ';
}


struct _TeletextDecoder{
    const ConsoleInterface* cons_vt;
    Console* cons;
    union{
        TeletextState state;
        unsigned int attr:CONSOLE_BITS_ATTR;
    };
    int x;
    int y;
};


TeletextDecoder* teletext_decoder_new(const ConsoleInterface* cons_vt,Console* cons){
    TeletextDecoder* dec = (TeletextDecoder*)malloc(sizeof(TeletextDecoder));
    dec->cons_vt = cons_vt;
    dec->cons = cons;
    dec->x = 0;
    dec->y = 0;
    return dec;
}


void teletext_decoder_free(TeletextDecoder* dec){
    free(dec);
}


void teletext_decoder_goto(TeletextDecoder* dec,int x, int y){
    dec->x = x;
    dec->y = y;
}


void teletext_decoder_write(TeletextDecoder* dec,int ch){
    if(0 == dec->x)
        teletext_reset(&dec->state);
    ch = teletext_update(&dec->state,ch);
    dec->cons_vt->charxya(dec->cons,dec->x,dec->y,dec->attr,ch);
    if(++dec->x >= TT_SCREEN_WIDTH){
        dec->x = 0;
        if(++dec->y >= TT_SCREEN_HEIGHT)
            dec->y = 0;
    }
}
