/*
 * Copyright (C) 2018 - 2019  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <byteswap.h>
#include <cairo.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>                     /* memcpy() */
#include <sys/stat.h>

#include "debug.h"
#include "flatten.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


const char* showGimpLayerModeEffects(GimpLayerModeEffects x){
    return "";
}


const char* showGimpImageBaseType(GimpImageBaseType x){
    return "";
}


const char* showGimpImageType(GimpImageType x){
    return "";
}


const char* showPropType(PropType x){
    return "";
}


const char* showXcfCompressionType(XcfCompressionType x){
    return "";
}


static void vFatalGeneric(int status,const char* format,va_list args){
#if 0
    if(format){
        char* s;
        if(*format == '!'){
            s = u_strdupvf(format+1,args);
            DEBUG_p("%s: %s",s,strerror(errno));
        }else{
            s = u_strdupvf(format+1,args);
            /DEBUG_p("%s",s);
        }
        u_free(s);
    }
#endif
}


void FatalGeneric(int status,const char* format,...){
    va_list v;
    va_start(v,format);
    vFatalGeneric(status,format,v);
}


void FatalUnexpected(const char* format,...){
    va_list v;
    va_start(v,format);
    vFatalGeneric(127,format,v);
}


void FatalBadXCF(const char* format,...){
#ifdef USE_DEBUG
    va_list v;
    va_start(v,format);
    DEBUG_p("corrupted or malformed XCF file");
    vFatalGeneric(125,format,v);
#endif
}


void xcfCheckspace(struct xcfGlobal *global,uint32_t addr,int spaceafter,const char *format,...){
    if( global->xcf_length < spaceafter || addr > global->xcf_length - spaceafter ) {
#ifdef USE_DEBUG
        va_list v;
        DEBUG_p("corrupted or truncated XCF file");
        DEBUG_p("(0x%" PRIXPTR " bytes): ",(uintptr_t)global->xcf_length);
        va_start(v,format);
        vFatalGeneric(125,format,v);
#endif
    }
}


void FatalUnsupportedXCF(const char* format,...){
    va_list v;
    va_start(v,format);
    DEBUG_p("the image contains features not understood by this program:");
    vFatalGeneric(123,format,v) ;
}


void* xcfmalloc(size_t size){
    return malloc(size);
}


void xcffree(void* block,struct xcfGlobal* global){
    if(global->xcf_file &&
      (uint8_t*)block >= global->xcf_file &&
      (uint8_t*)block < global->xcf_file + global->xcf_length )
        ;
    else
        free(block);
}


static enum out_color_mode guessIndexed(struct FlattenSpec* spec,rgba** pixels,void* userdata){
    DEBUG_p("begin");
    return COLOR_BY_CONTENTS;
}


static const char **
_xcf_get_mime_types()
{
    static const char *mime_types[] = {
        "image/xcf",
        "image/x-xcf",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_xcf_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_xcf_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


typedef struct{
    uint32_t *dst;
    int rowstride;
    int line;
}XcfData;
static void _xcf_create_from_file_cb(unsigned int w,rgba* pixels,void* userdata){
    XcfData* d = (XcfData*)userdata;
    int i;
    for(i=0;i<w;++i)
        d->dst[i] = bswap_32(pixels[i]);
    d->dst += d->rowstride;
    ++d->line;
}
static int
_xcf_create_from_file(abydos_plugin_handle_t *h, const char* filename)
{
    struct stat st;
    struct xcfGlobal global;
    struct xcfImage xcf;
    struct FlattenSpec flatspec;
    XcfData d;

    if(stat(filename,&st)<0){
        h->info->error = strerror(errno);
        return -1;
    }

    memset(&global,0,sizeof(struct xcfGlobal));

    read_or_mmap_xcf(&global,filename,NULL);
    getBasicXcfInfo(&global,&xcf);
    initColormap(&global,&xcf);

    init_flatspec(&flatspec);
    complete_flatspec(&global,&xcf,&flatspec,guessIndexed,NULL);
    DEBUG_p("src size: %dx%d",flatspec.dim.width,flatspec.dim.height);

    h->info->width = flatspec.dim.width;
    h->info->height = flatspec.dim.height;

    h->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,flatspec.dim.width,flatspec.dim.height);
    DEBUG_p("dst size: %dx%d",cairo_image_surface_get_width(h->surface),cairo_image_surface_get_height(h->surface));
    d.dst = (uint32_t*)cairo_image_surface_get_data(h->surface);
    d.rowstride = cairo_image_surface_get_stride(h->surface)/sizeof(uint32_t);
    d.line = 0;

    cairo_surface_flush(h->surface);
    flattenIncrementally(&global,&xcf,&flatspec,_xcf_create_from_file_cb,&d);
    cairo_surface_mark_dirty(h->surface);
    return 0;
}


cairo_surface_t *
_xcf_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int variant,double time,int frame){
    /* FIXME!! render the visible layers */
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _xcf_get_mime_types,
        .new = _xcf_new,
        .free = _xcf_free,
        .create_from_file = _xcf_create_from_file,
        .get_image_surface = _xcf_get_image_surface
    };
    return &p;
}
