/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#include "dec_huff.h"
#include "dec_huffman.h"
#include "streami.h"
#include "streamo.h"


HUFFDecompressor::HUFFDecompressor(const Buffer &packedData) :
	XPKDecompressor(packedData)
{
	if (packedData.size()<6)
		throw Decompressor::InvalidFormatError();
	// version: only 0 is defined
	uint16_t ver=packedData.readBE16(0);
	if (ver) throw Decompressor::InvalidFormatError();
	// password: we do not support it...
	uint32_t pwd=packedData.readBE32(2);
	if (pwd!=0xabadcafe) throw Decompressor::InvalidFormatError();
}

void HUFFDecompressor::decompressImpl(Buffer &rawData,const Buffer &previousData,bool verify)
{
	ForwardInputStream inputStream(_packedData,6,_packedData.size());
	MSBBitReader<ForwardInputStream> bitReader(inputStream);
	auto readBit=[&]()->uint32_t
	{
		return bitReader.readBits8(1);
	};
	auto readByte=[&]()->uint8_t
	{
		return inputStream.readByte();
	};

	ForwardOutputStream outputStream(rawData,0,rawData.size());

	HuffmanDecoder<uint32_t> decoder;
	for (uint32_t i=0;i<256;i++)
	{
		uint8_t codeBits=readByte()+1;
		if (!codeBits) continue;
		if (codeBits>32) throw Decompressor::DecompressionError();
		uint32_t code=0;
		int32_t shift=-codeBits;
		for (uint32_t j=0;j<codeBits;j+=8)
		{
			code=(code<<8)|readByte();
			shift+=8;
		}
		code=(code>>shift)&((1<<codeBits)-1);
		decoder.insert(HuffmanCode<uint32_t>{codeBits,code,i});
	}

	while (!outputStream.eof())
		outputStream.writeByte(decoder.decode(readBit));
}
