/*
 * Copyright (C) 2020 - 2023  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <float.h>                      /* DBL_MAX */
#include <stddef.h>                     /* size_t */
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcmp(), memcpy() */

#include "debug.h"
#include "nile.h"


#define CHECK_MISMATCH(a,b) (((a)>0 && (b)>0 && (a) != (b)) || ((a)<=0 && (b)<=0))


static int
_calc_stride(int bytes, int stride)
{
    if(stride>0)
        return stride;
    switch(stride){
    case 0:
    case NILE_STRIDE_ALIGN8:
        break;
    case NILE_STRIDE_ALIGN16:
        if(bytes&1)
            bytes += 1;
        break;
    case NILE_STRIDE_ALIGN32:
        if(bytes&3)
            bytes += 4 - (bytes&3);
        break;
    case NILE_STRIDE_ALIGN64:
        if(bytes&7)
            bytes += 8 - (bytes&7);
        break;
    }
    return bytes;
}


static int
_calc_stride_bits(int bits, int stride)
{
    return _calc_stride((bits-1)/8+1, stride);
}


nile_t *
nile_init(nile_t *ni)
{
    memset(ni,0,sizeof(nile_t));
    return ni;
}


void
nile_done(nile_t *ni)
{
    if(ni->pixels_allocation)
        free(ni->pixels);
    if(ni->palette_allocation)
        free(ni->palette);
}


nile_t *
nile_new()
{
    return (nile_t*)calloc(1,sizeof(nile_t));
}


void
nile_free(nile_t* ni)
{
    nile_done(ni);
    free(ni);
}



const uint8_t *
nile_get_color(const nile_t *ni, int x, int y)
{
    int stride;
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_BITPLANES:
        /* FIXME!! */
        break;
    case NILE_TYPE_INDEXED:
        stride = nile_get_stride(ni);
        return ni->palette + ni->pixels[x + y*stride] * ni->bpc;
    case NILE_TYPE_DIRECT:
        stride = nile_get_stride(ni);
        return ni->pixels + x*ni->bpc + y*stride;
    }
    return NULL;
}



void nile_drop_pixels(nile_t *ni){
    if(ni->pixels_allocation){
        free(ni->pixels);
        ni->pixels_allocation = 0;
    }
    ni->pixels = NULL;
    if(ni->stride > 0)
        ni->stride = 0;
}


static void
_link_bitplanes(nile_t *ni)
{
    int offset = nile_get_stride(ni) * ni->height;
    int i;
    ni->bitplanes[0] = ni->pixels + ni->depth * sizeof(uint8_t*);
    for(i=1;i<ni->depth;++i)
        ni->bitplanes[i] = ni->bitplanes[i-1] + offset;
}


int
nile_ensure_bitplanes(nile_t *ni,int width, int height, int depth, int deep)
{
#if 1
    int stride;
    if(NILE_TYPE_BITPLANES == ni->type && ni->bitplanes && ni->width >= width && ni->height >= height && ni->depth >= depth)
        return 0;
    nile_drop_pixels(ni);
    stride = deep ? _calc_stride_bits(width, ni->stride) : 0;
    ni->type = NILE_TYPE_BITPLANES;
    ni->pixels_allocation = depth * sizeof(uint8_t*) + (size_t)stride * height * depth;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
    ni->depth = depth;
    if(deep)
        _link_bitplanes(ni);
#else
    int stride = (width-1)/8+1;
    int size = stride * height * planes;
    int i;
    ni->width = width;
    ni->height = height;
    ni->planes = planes;
    ni->bitplanes = malloc(ni->planes * sizeof(uint8_t*));
    ni->bitplanes[0] = malloc(size);
    for(i=1;i<ni->planes;++i)
        ni->bitplanes[i] = ni->bitplanes[i-1] + stride*ni->height;
#endif
    return 0;
}


int
nile_ensure_indexed(nile_t *ni,int width, int height){
#if 1
    int stride;
    if(NILE_TYPE_INDEXED == ni->type && ni->pixels && ni->width >= width && ni->height >= height)
        return 0;
    nile_drop_pixels(ni);
    stride = _calc_stride(width, ni->stride);
    ni->type = NILE_TYPE_INDEXED;
    ni->pixels_allocation = (size_t)stride * height;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
#else
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(colors > 256)
        return -1;
    if(ni->colors){
        if(ni->colors < colors)
            return -1;
    }else{
        ni->colors = colors;
    }
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    ni->indexed = malloc(ni->width*ni->height);
#endif
    return 0;
}


int
nile_ensure_direct(nile_t *ni,int width, int height, int bpc)
{
#if 1
    if(NILE_TYPE_DIRECT == ni->type && ni->pixels && ni->width >= width && ni->height && ni->bpc >= bpc)
        return 0;
    nile_drop_pixels(ni);
    ni->type = NILE_TYPE_DIRECT;
    ni->stride = _calc_stride(width * bpc, 0);
    ni->pixels_allocation = ni->stride * height;
    ni->pixels = malloc(ni->pixels_allocation);
    ni->width = width;
    ni->height = height;
    ni->bpc = bpc;
#else
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    if(bpc>0)
        ni->bpc = bpc;
    ni->planar = malloc(ni->width * ni->height * ni->bpc);
#endif
    return 0;
}


#if 0
void
nile_drop_planar(nile_t *ni)
{
    if(ni->planar){
        free(ni->planar);
        ni->planar = NULL;
    }
}
#endif


#if 0
void
nile_drop_non_planar(nile_t *ni)
{
    if(ni->bitplanes){
        free(ni->bitplanes[0]);
        free(ni->bitplanes);
        ni->bitplanes = NULL;
    }
    nile_drop_indexed(ni);
}
#endif


void
nile_drop_palette(nile_t *ni)
{
    if(ni->palette_allocation){
        free(ni->palette);
        ni->palette_allocation = 0;
    }
}


int
nile_ensure_palette(nile_t *ni,int colors, int bpc)
{
#if 1
    if(ni->palette && ni->colors >= colors && ni->bpc >= bpc)
        return 0;
    if(ni->palette_allocation){
        if(ni->palette_allocation < colors * bpc){
            ni->palette_allocation = colors * bpc;
            ni->palette = realloc(ni->palette, ni->palette_allocation);
        }
    }else{
        ni->palette_allocation = colors * bpc;
        ni->palette = malloc(ni->palette_allocation);
    }
    ni->colors = colors;
    ni->bpc = bpc;
#else
    if(CHECK_MISMATCH(ni->bpc, bpc))
        return -1;
    if(colors>0)
        ni->colors = colors;
    if(ni->colors <= 0)
        return -1;
    if(bpc>0)
        ni->bpc = bpc;
    if(ni->bpc <= 0)
        return -1;
    if(ni->palette){
        ni->palette = realloc(ni->palette, ni->colors * ni->bpc);
    }else{
        ni->palette = (uint8_t*)malloc(ni->colors * ni->bpc);
    }
#endif
    return 0;
}


#if 0
int
nile_get_bitplanes_stride(const nile_t *ni)
{
    int bytes = (ni->width-1)/8+1;
    return _calc_stride(bytes, ni->stride);
}


int
nile_get_indexed_stride(const nile_t *ni)
{
    return _calc_stride(ni->width, ni->stride);
}


int
nile_get_planar_stride(const nile_t *ni)
{
    return _calc_stride(ni->width*ni->bpc, ni->stride);
}
#endif


int
nile_get_stride(const nile_t *ni)
{
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_BITPLANES:
        return _calc_stride_bits(ni->width, ni->stride);
    case NILE_TYPE_INDEXED:
        return _calc_stride(ni->width, ni->stride);
    case NILE_TYPE_DIRECT:
        return _calc_stride(ni->width * ni->bpc, ni->stride);
    }
    return 0;
}


int
nile_get_skip(const nile_t *ni)
{
    switch(ni->type){
    case NILE_TYPE_NONE:
        break;
    case NILE_TYPE_BITPLANES:
        {
            int bytes = (ni->width-1)/8+1;
            return _calc_stride(bytes, ni->stride) - bytes;
        }
    case NILE_TYPE_INDEXED:
        return _calc_stride(ni->width, ni->stride) - ni->width;
    case NILE_TYPE_DIRECT:
        return _calc_stride(ni->width * ni->bpc, ni->stride) - ni->width * ni->bpc;
    }
    return 0;
}


#if 0
const uint8_t *
nile_planar_from_indexed(nile_t *ni){
    uint8_t *src = ni->indexed;
    int src_skip = nile_get_indexed_skip(ni);
    uint8_t *dst = ni->planar;
    int dst_skip = nile_get_planar_skip(ni);
    int y;
    for(y=0;y<ni->height;++y){
        int x;
        for(x=0;x<ni->width;++x){
            memcpy(dst, &ni->palette[(*src++)*ni->bpc], ni->bpc);
            dst += ni->bpc;
        }
        src += src_skip;
        dst += dst_skip;
    }
    return ni->planar;
}
#endif


double
nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2)
{
    int r1_o = NILE_OFFSET_CHANNEL(format1,0);
    int g1_o = NILE_OFFSET_CHANNEL(format1,1);
    int b1_o = NILE_OFFSET_CHANNEL(format1,2);
    int r2_o = NILE_OFFSET_CHANNEL(format2,0);
    int g2_o = NILE_OFFSET_CHANNEL(format2,1);
    int b2_o = NILE_OFFSET_CHANNEL(format2,2);
    if(NILE_GET_COLORSPACE(format1) != NILE_GET_COLORSPACE(format2))
        return DBL_MAX;
    switch(NILE_GET_COLORSPACE(format1)){
    case NILE_COLORSPACE_RGB:
        {
            double r = color1[r1_o]-color2[r2_o];
            double g = color1[g1_o]-color2[g2_o];
            double b = color1[b1_o]-color2[b2_o];
#if 1
            return 2.5*r*r + 4*g*g + 2.5*b*b;
#else
            return r*r + g*g + b*b;
#endif
        }
    default:
        break;
    }
    return DBL_MAX;
}


int
nile_best_color_index(const uint8_t *p, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format)
{
    double best_value = DBL_MAX;
    int best_index = -1;
    int i;
    for(i=0;i<colors;++i){
        double value = nile_color_variance(p, palette_format, color, color_format);
        if(value < best_value){
            best_value = value;
            best_index = i;
        }
        p += bpc;
    }
    return best_index;
}


static void
_blit_bytes(uint8_t *dst,
            int dst_stride,
            const uint8_t *src,
            int src_stride,
            int width,
            int height,
            int flags)
{
    if(flags & NILE_FLIPPED){
        src += src_stride * (height-1);
        src_stride *= -1;
    }
    while(height-->0){
        memcpy(dst, src, width);
        src += src_stride;
        dst += dst_stride;
    }
}


void
nile_blit_indexed(uint8_t *dst,
                  int dst_stride,
                  const uint8_t *src,
                  int src_stride,
                  int width,
                  int height,
                  int flags)
{
    _blit_bytes(dst, dst_stride, src, src_stride, width, height, flags);
}


static void
_blit_bitplanes_from_palette(nile_blit_t *blit, uint8_t *dst, uint8_t **src, off_t offs, int count)
{
    while(count>0){
        uint8_t b[8];
        int plane;
        int i;
        memset(b,0,8);
        for(plane=0;plane<blit->src_depth;++plane){
            for(i=0;i<8;++i)
                if(src[plane][offs] & (128>>i))
                    b[i] |= 1<<plane;
        }
        for(i=0;i<8;++i){
            if(count<=0)
                break;
            memcpy(dst, blit->palette + b[i]*blit->dst_bpc, blit->dst_bpc);
            dst += blit->dst_bpc;
            --count;
        }
        ++offs;
    }
}


static void
_blit_packed_from_palette(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    int ppb = 8 / blit->src_depth; /* Pixels per byte */
    int mask = 0xff >> (8-blit->src_depth);
    int i;
    for(i=0;i<count;++i){
        int pix;
        if(0 == i % ppb)
            pix = *src++;
        pix <<= blit->src_depth;
        memcpy(dst, blit->palette + ((pix>>8) & mask) * blit->dst_bpc, blit->dst_bpc);
        dst += blit->dst_bpc;
    }
}


static void
_blit_from_palette(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        memcpy(dst, blit->palette + (*src++) * blit->src_bpc, blit->src_bpc);
        dst += blit->dst_bpc;
    }
}

static void
_blit_copy(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    memcpy(dst, src, count*blit->src_bpc);
}

static void
_blit_8i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        dst[blit->dst_o[4]] = src[blit->src_o[4]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_fill(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = src[blit->src_o[0]];
        dst[blit->dst_o[1]] = src[blit->src_o[1]];
        dst[blit->dst_o[2]] = src[blit->src_o[2]];
        dst[blit->dst_o[4]] = 255;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int a = src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a*src[blit->src_o[0]]/255;
        dst[blit->dst_o[1]] = a*src[blit->src_o[1]]/255;
        dst[blit->dst_o[2]] = a*src[blit->src_o[2]]/255;
        dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_unmultiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int a = src[blit->src_o[4]];
        dst[blit->dst_o[0]] = 255*src[blit->src_o[0]]/a;
        dst[blit->dst_o[1]] = 255*src[blit->src_o[1]]/a;
        dst[blit->dst_o[2]] = 255*src[blit->src_o[2]]/a;
        dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_6i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        int v;
        v = src[blit->src_o[0]];
        dst[blit->dst_o[0]] = v << 2 | v >> 4;
        v = src[blit->src_o[1]];
        dst[blit->dst_o[1]] = v << 2 | v >> 4;
        v = src[blit->src_o[2]];
        dst[blit->dst_o[2]] = v << 2 | v >> 4;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_16i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] >> 8;
        dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] >> 8;
        dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] >> 8;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_16i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] >> 8;
        dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] >> 8;
        dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] >> 8;
        dst[blit->dst_o[4]] = *(uint16_t*)&src[blit->src_o[4]] >> 8;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3_from_32f3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]];
        dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]];
        dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_32f3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]];
        dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]];
        dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]];
        dst[blit->dst_o[4]] = *(float*)&src[blit->src_o[4]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}

static void
_blit_8i3a_from_32f3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = *(float*)&src[blit->src_o[4]];
        dst[blit->dst_o[0]] = a * *(float*)&src[blit->src_o[0]]/255;
        dst[blit->dst_o[1]] = a * *(float*)&src[blit->src_o[1]]/255;
        dst[blit->dst_o[2]] = a * *(float*)&src[blit->src_o[2]]/255;
        dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static void
_blit_32f3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(float*)&src[blit->src_o[0]];
        *(float*)&dst[blit->dst_o[1]] = *(float*)&src[blit->src_o[1]];
        *(float*)&dst[blit->dst_o[2]] = *(float*)&src[blit->src_o[2]];
        *(float*)&dst[blit->dst_o[4]] = *(float*)&src[blit->src_o[4]];
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static void
_blit_32f3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = *(float*)&src[blit->src_o[4]];
        *(float*)&dst[blit->dst_o[0]] = a * *(float*)&src[blit->src_o[0]];
        *(float*)&dst[blit->dst_o[1]] = a * *(float*)&src[blit->src_o[1]];
        *(float*)&dst[blit->dst_o[2]] = a * *(float*)&src[blit->src_o[2]];
        *(float*)&dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static void
_blit_32f3a_from_16i3a(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[4]] = *(uint16_t*)&src[blit->src_o[4]] / UINT16_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static void
_blit_32f3_from_16i3(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        *(float*)&dst[blit->dst_o[0]] = *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static void
_blit_32f3a_from_16i3a_multiply(nile_blit_t *blit, uint8_t *dst, const uint8_t *src, int count)
{
    while(count-->0){
        float a = (float)*(uint16_t*)&src[blit->src_o[4]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[0]] = a * *(uint16_t*)&src[blit->src_o[0]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[1]] = a * *(uint16_t*)&src[blit->src_o[1]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[2]] = a * *(uint16_t*)&src[blit->src_o[2]] / UINT16_MAX;
        *(float*)&dst[blit->dst_o[4]] = a;
        src += blit->src_bpc;
        dst += blit->dst_bpc;
    }
}


static int
_setup_palette(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(dst_bpc != src_bpc || dst_format != src_format){
        if(colors > 256){
            DEBUG_p("more that 256 colors is not supported");
            return -1;
        }
        nile_copy_colors_rgb(blit->tmp_palette, dst_bpc, dst_format, palette, src_bpc, src_format, colors);
        blit->palette = blit->tmp_palette;
    }else{
        blit->palette = palette;
    }
    return 0;
}


nile_blit_t *
nile_blit_init_bitplanes_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(_setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format) < 0)
        return NULL;
    blit->dst_bpc = dst_bpc;
    blit->src_depth = depth;
    blit->blit_bitplanes_func = _blit_bitplanes_from_palette;
    return blit;
}


nile_blit_t *
nile_blit_init_packed_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, int depth, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    if(_setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format) < 0)
        return NULL;
    blit->dst_bpc = dst_bpc;
    blit->src_depth = depth;
    blit->blit_func = _blit_packed_from_palette;
    return blit;
}


nile_blit_t *
nile_blit_init_indexed_rgb(nile_blit_t *blit, int dst_bpc, nile_format_t dst_format, const uint8_t *palette, int colors, int src_bpc, nile_format_t src_format)
{
    _setup_palette(blit, dst_bpc, dst_format, palette, colors, src_bpc, src_format);
    blit->dst_bpc = dst_bpc;
    blit->src_bpc = src_bpc;
    blit->blit_func = _blit_from_palette;
    return blit;
}


nile_blit_t *
nile_blit_init_direct_rgb(nile_blit_t *blit,int dst_bpc, nile_format_t dst_format, int src_bpc, nile_format_t src_format)
{
    blit->dst_bpc = dst_bpc;
    blit->src_bpc = src_bpc;
    if(dst_bpc == src_bpc && dst_format == src_format){
        blit->blit_func = _blit_copy;
    }else{
        blit->dst_o[0] = NILE_OFFSET_CHANNEL(dst_format,0);
        blit->dst_o[1] = NILE_OFFSET_CHANNEL(dst_format,1);
        blit->dst_o[2] = NILE_OFFSET_CHANNEL(dst_format,2);
        blit->dst_o[4] = NILE_OFFSET_ALPHA(dst_format);
        if(NILE_FORMAT_r6g6b6 == src_format){
            blit->src_o[0] = 0;
            blit->src_o[1] = 1;
            blit->src_o[2] = 2;
            blit->blit_func = _blit_8i3_from_6i3;
            return blit;
        }
        blit->src_o[0] = NILE_OFFSET_CHANNEL(src_format,0);
        blit->src_o[1] = NILE_OFFSET_CHANNEL(src_format,1);
        blit->src_o[2] = NILE_OFFSET_CHANNEL(src_format,2);
        blit->src_o[4] = NILE_OFFSET_ALPHA(src_format);
        switch(NILE_GET_BYTES_PER_CHANNEL(dst_format)){
        case 1:
            switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
            case 1:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_8i3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_8i3a_multiply;
                    }else{
                        blit->blit_func = _blit_8i3a_unmultiply;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    blit->blit_func = _blit_8i3a_fill;
                }else{
                    blit->blit_func = _blit_8i3;
                }
                break;
            case 2:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_8i3a_from_16i3a;
                    }else{
                        return NULL;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    /* FIXME!! */
                    /* blit->blit_func = _blit_8i3a_from_16i3_fill; */
                    return NULL;
                }else{
                    blit->blit_func = _blit_8i3_from_16i3;
                }
                break;
            case 4:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_8i3a_from_32f3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_8i3a_from_32f3a_multiply;
                    }else{
                        /* blit->blit_func = _blit_8i3a_from_32f3a_multiply */
                        return NULL;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    /* blit->blit_func = _blit_8i3a_from_32f3_fill; */
                    return NULL;
                }else{
                    blit->blit_func = _blit_8i3_from_32f3;
                }
                break;
            default:
                return NULL;
            }
            break;
        case 4: /* assume float */
            switch(NILE_GET_BYTES_PER_CHANNEL(src_format)){
            case 2:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Both have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_32f3a_from_16i3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_32f3a_from_16i3a_multiply;
                    }else{
                        return NULL;
                    }
                }else if(NILE_HAS_ALPHA(dst_format)){
                    return NULL;
                }else{
                    blit->blit_func = _blit_32f3_from_16i3;
                }
                break;
            case 4:
                if(NILE_HAS_ALPHA(src_format) && NILE_HAS_ALPHA(dst_format)){
                    /* Bath have alpha channel */
                    if((dst_format & NILE_PREMULTIPLIED) == (src_format & NILE_PREMULTIPLIED)){
                        blit->blit_func = _blit_32f3a;
                    }else if(dst_format & NILE_PREMULTIPLIED){
                        blit->blit_func = _blit_32f3a_multiply;
                    }else{
                        return NULL;
                    }
                }else{
                    return NULL;
                }
                break;
            default:
                return NULL;
            }
            break;
        default:
            return NULL;
        }
    }
    return blit;
}


static void
_blit_bitplanes_from_indexed_line(uint8_t **dst, off_t offs, int depth, const uint8_t *src, int count)
{
    while(count>0){
        int plane;
        for(plane=0;plane<depth;++plane){
            uint8_t b = 0;
            int i;
            for(i=0;i<8;++i){
                if(i >= count)
                    break;
                if(src[i] & (1<<plane))
                    b |= 128>>i;
            }
            dst[plane][offs] = b;
        }
        ++offs;
        src += 8;
        count -= 8;
    }
}


static void
_blit_indexed_from_bitplanes_line(uint8_t *dst, uint8_t **src, off_t offs, int depth, int count)
{
    while(count>0){
        uint8_t b[8];
        int plane;
        int i;
        memset(b,0,8);
        for(plane=0;plane<depth;++plane){
            for(i=0;i<8;++i)
                if(src[plane][offs] & (128>>i))
                    b[i] |= 1<<plane;
        }
        for(i=0;i<8;++i){
            if(count<=0)
                break;
            *dst++ = b[i];
            --count;
        }
        ++offs;
    }
}


void
nile_blit(nile_blit_t *blit, uint8_t *dst, int dst_stride, const uint8_t *src, int src_stride, int width, int height)
{
    while(height-->0){
        nile_blit_line(blit, dst, src, width);
        dst += dst_stride;
        src += src_stride;
    }
}


void
nile_blit_bitplanes_from_indexed(uint8_t **dst, int dst_stride, int depth, const uint8_t *src, int src_stride, int width, int height)
{
    off_t offs = 0;
    while(height-->0){
        _blit_bitplanes_from_indexed_line(dst, offs, depth, src, width);
        offs += dst_stride;
        src += src_stride;
    }
}


void
nile_blit_indexed_from_bitplanes(uint8_t *dst, int dst_stride, uint8_t **src, int src_stride, int depth, int width, int height)
{
    off_t offs = 0;
    while(height-->0){
        _blit_indexed_from_bitplanes_line(dst, src, offs, depth, width);
        dst += dst_stride;
        offs += src_stride;
    }
}


void
nile_blit_direct_rgb(uint8_t *dst,
                     int dst_stride,
                     int dst_bpc,
                     nile_format_t dst_format,
                     const uint8_t *src,
                     int src_stride,
                     int src_bpc,
                     nile_format_t src_format,
                     int width,
                     int height,
                     int flags)
{
    nile_blit_t blit;
    nile_blit_init_direct_rgb(&blit, dst_bpc, dst_format, src_bpc, src_format);
    if(flags & NILE_FLIPPED){
        src += src_stride * (height-1);
        src_stride *= -1;
    }
    while(height-->0){
        nile_blit_line(&blit, dst, src, width);
        dst += dst_stride;
        src += src_stride;
    }
}


void
nile_blit_direct_from_bitplanes_rgb(uint8_t *dst,
                                    int dst_stride,
                                    int dst_bpc,
                                    nile_format_t dst_format,
                                    uint8_t **src,
                                    int src_stride,
                                    int depth,
                                    const uint8_t *palette,
                                    int colors,
                                    int src_bpc,
                                    nile_format_t src_format,
                                    int width,
                                    int height,
                                    int flags)
{
    nile_blit_t blit;
    off_t offs = 0;
    nile_blit_init_bitplanes_rgb(&blit, dst_bpc, dst_format, depth, palette, colors, src_bpc, src_format);
    if(flags & NILE_FLIPPED){
        offs += src_stride * (height-1);
        src_stride *= -1;
    }
    while(height-->0){
        nile_blit_bitplanes_line(&blit, dst, src, offs, width);
        dst += dst_stride;
        offs += src_stride;
    }
}


void
nile_blit_direct_from_indexed_rgb(uint8_t *dst,
                                  int dst_stride,
                                  int dst_bpc,
                                  nile_format_t dst_format,
                                  const uint8_t *src,
                                  int src_stride,
                                  const uint8_t *palette,
                                  int colors,
                                  int src_bpc,
                                  nile_format_t src_format,
                                  int width,
                                  int height,
                                  int flags)
{
    uint8_t tmp_palette[NILE_MAX_BYTES_PER_COLOR * 256];
    int dst_skip = dst_stride - width*dst_bpc;
    int src_skip = src_stride - width;
    if(dst_bpc != src_bpc || dst_format != src_format){
        nile_copy_colors_rgb(tmp_palette, dst_bpc, dst_format, palette, src_bpc, src_format, colors);
        palette = tmp_palette;
    }
    if(flags & NILE_FLIPPED){
        src += src_stride * (height-1);
        src_skip -= src_stride*2;
    }
    while(height-->0){
        int x = width;
        while(x-->0){
            int idx = *src++;
            memcpy(dst, palette + idx*dst_bpc, dst_bpc);
            dst += dst_bpc;
        }
        dst += dst_skip;
        src += src_skip;
    }
}


void
nile_copy_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count)
{
    nile_blit_direct_rgb(dst, 0, dst_bpc, dst_format, src, 0, src_bpc, src_format, count, 1, 0);
}



void
nile_copy(nile_t *dst, const nile_t *src, int deep)
{
    memcpy(dst, src, sizeof(nile_t));
    if(deep){
        if(dst->pixels_allocation){
            dst->pixels = malloc(dst->pixels_allocation);
            memcpy(dst->pixels, src->pixels, dst->pixels_allocation);
            if(NILE_TYPE_BITPLANES == dst->type)
                _link_bitplanes(dst);
        }
        if(dst->palette_allocation){
            dst->palette = malloc(dst->palette_allocation);
            memcpy(dst->palette, src->palette, dst->palette_allocation);
        }
    }else{
        memcpy(dst, src, sizeof(nile_t));
        dst->pixels_allocation = dst->palette_allocation = 0;
    }
}


int
nile_copy_bitplanes_from_indexed(nile_t *dst,const nile_t *src, int deep)
{
    memcpy(dst, src, sizeof(nile_t));
    if(deep){
        if(dst->palette_allocation){
            dst->palette = malloc(dst->palette_allocation);
            memcpy(dst->palette, src->palette, dst->palette_allocation);
        }
    }else{
        dst->palette_allocation = 0;
    }
    dst->pixels_allocation = 0;
    if(!dst->depth){
        dst->depth = 1;
        while(1<<dst->depth < dst->colors)
            ++dst->depth;
    }
    nile_ensure_bitplanes(dst, src->width, src->height, dst->depth, 1);
    nile_blit_bitplanes_from_indexed(dst->bitplanes, nile_get_stride(dst), dst->depth,
                                     src->pixels, nile_get_stride(src),
                                     src->width, src->height);
    return 0;
}


int nile_copy_indexed_from_bitplanes(nile_t *dst, const nile_t *src, int deep){
    memcpy(dst, src, sizeof(nile_t));
    if(deep){
        if(dst->palette_allocation){
            dst->palette = malloc(dst->palette_allocation);
            memcpy(dst->palette, src->palette, dst->palette_allocation);
        }
    }else{
        dst->palette_allocation = 0;
    }
    dst->pixels_allocation = 0;
    nile_ensure_indexed(dst, src->width, src->height);
    if(!dst->colors)
        dst->colors = 1<<src->depth;
    nile_blit_indexed_from_bitplanes(dst->pixels, nile_get_stride(dst),
                                     src->bitplanes, nile_get_stride(src), src->depth,
                                     src->width, src->height);
    return 0;
}


int nile_copy_direct_from_bitplanes(nile_t *dst, const nile_t *src){
    nile_ensure_direct(dst, src->width,  src->height, src->bpc);
    dst->format = src->format;
    nile_blit_direct_from_bitplanes_rgb(dst->pixels, nile_get_stride(dst),
                                        dst->bpc, dst->format,
                                        src->bitplanes, nile_get_stride(src), src->depth,
                                        src->palette, src->colors,
                                        src->bpc, src->format,
                                        src->width, src->height,
                                        0);
    return 0;
}


int nile_copy_direct_from_indexed(nile_t *dst,const nile_t *src){
    nile_ensure_direct(dst, src->width,  src->height, src->bpc);
    dst->format = src->format;
    nile_blit_direct_from_indexed_rgb(dst->pixels, nile_get_stride(dst),
                                      dst->bpc, dst->format,
                                      src->pixels, nile_get_stride(src),
                                      src->palette, src->colors,
                                      src->bpc, src->format,
                                      src->width, src->height,
                                      0);
    return 0;
}


void
nile_move(nile_t *dst, nile_t *src)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    if(dst->palette_allocation)
        free(dst->palette);
    memcpy(dst, src, sizeof(nile_t));
    src->pixels_allocation = src->palette_allocation = 0;
}


void
nile_copy_pixels(nile_t *dst, const nile_t *src, int deep)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    dst->type = src->type;
    if(deep && src->pixels_allocation){
        dst->pixels_allocation = src->pixels_allocation;
        dst->pixels = malloc(dst->pixels_allocation);
        memcpy(dst->pixels, src->pixels, dst->pixels_allocation);
        if(NILE_TYPE_BITPLANES == dst->type)
            _link_bitplanes(dst);
    }else{
        dst->pixels_allocation = 0;
        dst->pixels = src->pixels;
    }
    dst->width = src->width;
    dst->height = src->height;
    dst->stride = src->stride;
    if(NILE_TYPE_DIRECT == dst->type){
        dst->bpc = src->bpc;
        dst->format = src->format;
    }
}


int
nile_copy_pixels_bitplanes_from_indexed(nile_t *dst, const nile_t *src)
{
    if(dst->pixels_allocation){
        free(dst->pixels);
        dst->pixels_allocation = 0;
    }
    dst->width = src->width;
    dst->height = src->height;
    if(!dst->depth){
        dst->depth = 1;
        while(1<<dst->depth < dst->colors)
            ++dst->depth;
    }
    nile_ensure_bitplanes(dst, src->width, src->height, dst->depth, 1);
    nile_blit_bitplanes_from_indexed(dst->bitplanes, nile_get_stride(dst), dst->depth,
                                     src->pixels, nile_get_stride(src),
                                     src->width, src->height);
    return 0;
}


int nile_copy_pixels_indexed_from_bitplanes(nile_t *dst, const nile_t *src)
{
    if(dst->pixels_allocation){
        free(dst->pixels);
        dst->pixels_allocation = 0;
    }
    dst->width = src->width;
    dst->height = src->height;
    if(!dst->colors)
        dst->colors = 1<<src->depth;
    nile_ensure_indexed(dst, src->width, src->height);
    nile_blit_indexed_from_bitplanes(dst->pixels, nile_get_stride(dst),
                                     src->bitplanes, nile_get_stride(src), src->depth,
                                     src->width, src->height);
    return 0;
}


void
nile_move_pixels(nile_t *dst, nile_t *src)
{
    if(dst->pixels_allocation)
        free(dst->pixels);
    dst->type = src->type;
    dst->pixels_allocation = src->pixels_allocation;
    dst->pixels = src->pixels;
    dst->width = src->width;
    dst->height = src->height;
    dst->stride = src->stride;
    if(NILE_TYPE_BITPLANES == src->type){
        dst->depth = src->depth;
    }else if(NILE_TYPE_DIRECT == src->type){
        dst->bpc = src->bpc;
        dst->format = src->format;
    }
    src->pixels_allocation = 0;
}


void
nile_copy_palette(nile_t *dst, const nile_t *src, int deep)
{
    if(dst->palette_allocation)
        free(dst->palette);
    if(deep && src->palette_allocation){
        dst->palette_allocation = src->palette_allocation;
        dst->palette = malloc(dst->palette_allocation);
        memcpy(dst->palette, src->palette, dst->palette_allocation);
    }else{
        dst->palette_allocation = 0;
        dst->palette = src->palette;
    }
    dst->colors = src->colors;
    dst->bpc = src->bpc;
    dst->format = src->format;
}


void
nile_move_palette(nile_t *dst, nile_t *src)
{
    if(dst->palette_allocation)
        free(dst->palette);
    dst->palette_allocation = src->palette_allocation;
    dst->palette = src->palette;
    dst->colors = src->colors;
    src->palette_allocation = 0;
    dst->bpc = src->bpc;
    dst->format = src->format;
}
